###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import inspect
import six
from six import StringIO

from autobahn.wamp.interfaces import ISession, \
    IPublication, \
    IPublisher, \
    ISubscription, \
    ISubscriber, \
    ICaller, \
    IRegistration, \
    ITransportHandler

from autobahn import util
from autobahn import wamp
from autobahn.wamp import uri
from autobahn.wamp import message
from autobahn.wamp import types
from autobahn.wamp import role
from autobahn.wamp import exception
from autobahn.wamp.exception import ProtocolError, SessionNotReady
from autobahn.wamp.types import SessionDetails


def is_method_or_function(f):
    return inspect.ismethod(f) or inspect.isfunction(f)


class Endpoint:
    """
    """

    def __init__(self, obj, fn, procedure, options=None):
        self.obj = obj
        self.fn = fn
        self.procedure = procedure
        self.options = options


class Handler:
    """
    """

    def __init__(self, obj, fn, topic, details_arg=None):
        self.obj = obj
        self.fn = fn
        self.topic = topic
        self.details_arg = details_arg


class Publication:
    """
    Object representing a publication.
    This class implements :class:`autobahn.wamp.interfaces.IPublication`.
    """
    def __init__(self, publicationId):
        self.id = publicationId


IPublication.register(Publication)


class Subscription:
    """
    Object representing a subscription.
    This class implements :class:`autobahn.wamp.interfaces.ISubscription`.
    """
    def __init__(self, session, subscriptionId):
        self._session = session
        self.active = True
        self.id = subscriptionId

    def unsubscribe(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISubscription.unsubscribe`
        """
        return self._session._unsubscribe(self)


ISubscription.register(Subscription)


class Registration:
    """
    Object representing a registration.
    This class implements :class:`autobahn.wamp.interfaces.IRegistration`.
    """
    def __init__(self, session, registrationId):
        self._session = session
        self.active = True
        self.id = registrationId

    def unregister(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IRegistration.unregister`
        """
        return self._session._unregister(self)


IRegistration.register(Registration)


class BaseSession:
    """
    WAMP session base class.

    This class implements :class:`autobahn.wamp.interfaces.ISession`.
    """

    def __init__(self):
        """

        """
        # this is for library level debugging
        self.debug = False

        # this is for app level debugging. exceptions raised in user code
        # will get logged (that is, when invoking remoted procedures or
        # when invoking event handlers)
        self.debug_app = False

        # this is for marshalling traceback from exceptions thrown in user
        # code within WAMP error messages (that is, when invoking remoted
        # procedures)
        self.traceback_app = False

        # mapping of exception classes to WAMP error URIs
        self._ecls_to_uri_pat = {}

        # mapping of WAMP error URIs to exception classes
        self._uri_to_ecls = {}

        # session authentication information
        self._authid = None
        self._authrole = None
        self._authmethod = None
        self._authprovider = None

    def onConnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onConnect`
        """

    def onJoin(self, details):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onJoin`
        """

    def onLeave(self, details):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onLeave`
        """

    def onDisconnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onDisconnect`
        """

    def define(self, exception, error=None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.define`
        """
        if error is None:
            assert(hasattr(exception, '_wampuris'))
            self._ecls_to_uri_pat[exception] = exception._wampuris
            self._uri_to_ecls[exception._wampuris[0].uri()] = exception
        else:
            assert(not hasattr(exception, '_wampuris'))
            self._ecls_to_uri_pat[exception] = [uri.Pattern(six.u(error), uri.Pattern.URI_TARGET_HANDLER)]
            self._uri_to_ecls[six.u(error)] = exception

    def _message_from_exception(self, request_type, request, exc, tb=None):
        """
        Create a WAMP error message from an exception.

        :param request_type: The request type this WAMP error message is for.
        :type request_type: int
        :param request: The request ID this WAMP error message is for.
        :type request: int
        :param exc: The exception.
        :type exc: Instance of :class:`Exception` or subclass thereof.
        :param tb: Optional traceback. If present, it'll be included with the WAMP error message.
        :type tb: list or None
        """
        args = None
        if hasattr(exc, 'args'):
            args = list(exc.args)  # make sure tuples are made into lists

        kwargs = None
        if hasattr(exc, 'kwargs'):
            kwargs = exc.kwargs

        if tb:
            if kwargs:
                kwargs['traceback'] = tb
            else:
                kwargs = {'traceback': tb}

        if isinstance(exc, exception.ApplicationError):
            error = exc.error if type(exc.error) == six.text_type else six.u(exc.error)
        else:
            if exc.__class__ in self._ecls_to_uri_pat:
                error = self._ecls_to_uri_pat[exc.__class__][0]._uri
            else:
                error = u"wamp.error.runtime_error"

        msg = message.Error(request_type, request, error, args, kwargs)

        return msg

    def _exception_from_message(self, msg):
        """
        Create a user (or generic) exception from a WAMP error message.

        :param msg: A WAMP error message.
        :type msg: instance of :class:`autobahn.wamp.message.Error`
        """

        # FIXME:
        # 1. map to ecls based on error URI wildcard/prefix
        # 2. extract additional args/kwargs from error URI

        exc = None

        if msg.error in self._uri_to_ecls:
            ecls = self._uri_to_ecls[msg.error]
            try:
                # the following might fail, eg. TypeError when
                # signature of exception constructor is incompatible
                # with args/kwargs or when the exception constructor raises
                if msg.kwargs:
                    if msg.args:
                        exc = ecls(*msg.args, **msg.kwargs)
                    else:
                        exc = ecls(**msg.kwargs)
                else:
                    if msg.args:
                        exc = ecls(*msg.args)
                    else:
                        exc = ecls()
            except Exception:
                # FIXME: log e
                pass

        if not exc:
            # the following ctor never fails ..
            if msg.kwargs:
                if msg.args:
                    exc = exception.ApplicationError(msg.error, *msg.args, **msg.kwargs)
                else:
                    exc = exception.ApplicationError(msg.error, **msg.kwargs)
            else:
                if msg.args:
                    exc = exception.ApplicationError(msg.error, *msg.args)
                else:
                    exc = exception.ApplicationError(msg.error)

        return exc


ISession.register(BaseSession)


class ApplicationSession(BaseSession):
    """
    WAMP endpoint session. This class implements

    * :class:`autobahn.wamp.interfaces.IPublisher`
    * :class:`autobahn.wamp.interfaces.ISubscriber`
    * :class:`autobahn.wamp.interfaces.ICaller`
    * :class:`autobahn.wamp.interfaces.ICallee`
    * :class:`autobahn.wamp.interfaces.ITransportHandler`
    """

    def __init__(self, config=None):
        """
        Constructor.
        """
        BaseSession.__init__(self)
        self.config = config or types.ComponentConfig(realm=u"default")

        self._transport = None
        self._session_id = None
        self._realm = None

        self._session_id = None
        self._goodbye_sent = False
        self._transport_is_closing = False

        # outstanding requests
        self._publish_reqs = {}
        self._subscribe_reqs = {}
        self._unsubscribe_reqs = {}
        self._call_reqs = {}
        self._register_reqs = {}
        self._unregister_reqs = {}

        # subscriptions in place
        self._subscriptions = {}

        # registrations in place
        self._registrations = {}

        # incoming invocations
        self._invocations = {}

    def onOpen(self, transport):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onOpen`
        """
        self._transport = transport
        self.onConnect()

    def onConnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onConnect`
        """
        self.join(self.config.realm)

    def join(self, realm, authmethods=None, authid=None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.join`
        """
        if six.PY2 and type(realm) == str:
            realm = six.u(realm)

        if self._session_id:
            raise Exception("already joined")

        self._goodbye_sent = False

        roles = [
            role.RolePublisherFeatures(),
            role.RoleSubscriberFeatures(),
            role.RoleCallerFeatures(),
            role.RoleCalleeFeatures()
        ]

        msg = message.Hello(realm, roles, authmethods, authid)
        self._realm = realm
        self._transport.send(msg)

    def disconnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.disconnect`
        """
        if self._transport:
            self._transport.close()
        else:
            raise Exception("transport disconnected")

    def onMessage(self, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onMessage`
        """
        if self._session_id is None:

            # the first message must be WELCOME, ABORT or CHALLENGE ..
            if isinstance(msg, message.Welcome):
                self._session_id = msg.session

                details = SessionDetails(self._realm, self._session_id, msg.authid, msg.authrole, msg.authmethod)
                self._as_future(self.onJoin, details)

            elif isinstance(msg, message.Abort):

                # fire callback and close the transport
                self.onLeave(types.CloseDetails(msg.reason, msg.message))

            elif isinstance(msg, message.Challenge):

                challenge = types.Challenge(msg.method, msg.extra)
                d = self._as_future(self.onChallenge, challenge)

                def success(signature):
                    reply = message.Authenticate(signature)
                    self._transport.send(reply)

                def error(err):
                    reply = message.Abort(u"wamp.error.cannot_authenticate", u"{0}".format(err.value))
                    self._transport.send(reply)
                    # fire callback and close the transport
                    self.onLeave(types.CloseDetails(reply.reason, reply.message))

                self._add_future_callbacks(d, success, error)

            else:
                raise ProtocolError("Received {0} message, and session is not yet established".format(msg.__class__))

        else:

            if isinstance(msg, message.Goodbye):
                if not self._goodbye_sent:
                    # the peer wants to close: send GOODBYE reply
                    reply = message.Goodbye()
                    self._transport.send(reply)

                self._session_id = None

                # fire callback and close the transport
                self.onLeave(types.CloseDetails(msg.reason, msg.message))

            # consumer messages
            elif isinstance(msg, message.Event):

                if msg.subscription in self._subscriptions:

                    handler = self._subscriptions[msg.subscription]

                    if handler.details_arg:
                        if not msg.kwargs:
                            msg.kwargs = {}
                        msg.kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher)

                    try:
                        if handler.obj:
                            if msg.kwargs:
                                if msg.args:
                                    handler.fn(handler.obj, *msg.args, **msg.kwargs)
                                else:
                                    handler.fn(handler.obj, **msg.kwargs)
                            else:
                                if msg.args:
                                    handler.fn(handler.obj, *msg.args)
                                else:
                                    handler.fn(handler.obj)
                        else:
                            if msg.kwargs:
                                if msg.args:
                                    handler.fn(*msg.args, **msg.kwargs)
                                else:
                                    handler.fn(**msg.kwargs)
                            else:
                                if msg.args:
                                    handler.fn(*msg.args)
                                else:
                                    handler.fn()

                    except Exception as e:
                        if self.debug_app:
                            print("Failure while firing event handler {0} subscribed under '{1}' ({2}): {3}".format(handler.fn, handler.topic, msg.subscription, e))

                else:
                    raise ProtocolError("EVENT received for non-subscribed subscription ID {0}".format(msg.subscription))

            elif isinstance(msg, message.Published):

                if msg.request in self._publish_reqs:
                    d, opts = self._publish_reqs.pop(msg.request)
                    p = Publication(msg.publication)
                    self._resolve_future(d, p)
                else:
                    raise ProtocolError("PUBLISHED received for non-pending request ID {0}".format(msg.request))

            elif isinstance(msg, message.Subscribed):

                if msg.request in self._subscribe_reqs:
                    d, obj, fn, topic, options = self._subscribe_reqs.pop(msg.request)
                    if options:
                        self._subscriptions[msg.subscription] = Handler(obj, fn, topic, options.details_arg)
                    else:
                        self._subscriptions[msg.subscription] = Handler(obj, fn, topic)
                    s = Subscription(self, msg.subscription)
                    self._resolve_future(d, s)
                else:
                    raise ProtocolError("SUBSCRIBED received for non-pending request ID {0}".format(msg.request))

            elif isinstance(msg, message.Unsubscribed):

                if msg.request in self._unsubscribe_reqs:
                    d, subscription = self._unsubscribe_reqs.pop(msg.request)
                    if subscription.id in self._subscriptions:
                        del self._subscriptions[subscription.id]
                    subscription.active = False
                    self._resolve_future(d, None)
                else:
                    raise ProtocolError("UNSUBSCRIBED received for non-pending request ID {0}".format(msg.request))

            elif isinstance(msg, message.Result):

                if msg.request in self._call_reqs:

                    if msg.progress:

                        # progressive result
                        _, opts = self._call_reqs[msg.request]
                        if opts.onProgress:
                            try:
                                if msg.kwargs:
                                    if msg.args:
                                        opts.onProgress(*msg.args, **msg.kwargs)
                                    else:
                                        opts.onProgress(**msg.kwargs)
                                else:
                                    if msg.args:
                                        opts.onProgress(*msg.args)
                                    else:
                                        opts.onProgress()
                            except Exception as e:
                                # silently drop exceptions raised in progressive results handlers
                                if self.debug:
                                    print("Exception raised in progressive results handler: {0}".format(e))
                        else:
                            # silently ignore progressive results
                            pass
                    else:

                        # final result
                        d, opts = self._call_reqs.pop(msg.request)
                        if msg.kwargs:
                            if msg.args:
                                res = types.CallResult(*msg.args, **msg.kwargs)
                            else:
                                res = types.CallResult(**msg.kwargs)
                            self._resolve_future(d, res)
                        else:
                            if msg.args:
                                if len(msg.args) > 1:
                                    res = types.CallResult(*msg.args)
                                    self._resolve_future(d, res)
                                else:
                                    self._resolve_future(d, msg.args[0])
                            else:
                                self._resolve_future(d, None)
                else:
                    raise ProtocolError("RESULT received for non-pending request ID {0}".format(msg.request))

            elif isinstance(msg, message.Invocation):

                if msg.request in self._invocations:

                    raise ProtocolError("INVOCATION received for request ID {0} already invoked".format(msg.request))

                else:

                    if msg.registration not in self._registrations:

                        raise ProtocolError("INVOCATION received for non-registered registration ID {0}".format(msg.registration))

                    else:
                        endpoint = self._registrations[msg.registration]

                        if endpoint.options and endpoint.options.details_arg:

                            if not msg.kwargs:
                                msg.kwargs = {}

                            if msg.receive_progress:
                                def progress(*args, **kwargs):
                                    progress_msg = message.Yield(msg.request, args=args, kwargs=kwargs, progress=True)
                                    self._transport.send(progress_msg)
                            else:
                                progress = None

                            msg.kwargs[endpoint.options.details_arg] = types.CallDetails(progress, caller=msg.caller,
                                                                                         caller_transport=msg.caller_transport, authid=msg.authid, authrole=msg.authrole,
                                                                                         authmethod=msg.authmethod)

                        if endpoint.obj:
                            if msg.kwargs:
                                if msg.args:
                                    d = self._as_future(endpoint.fn, endpoint.obj, *msg.args, **msg.kwargs)
                                else:
                                    d = self._as_future(endpoint.fn, endpoint.obj, **msg.kwargs)
                            else:
                                if msg.args:
                                    d = self._as_future(endpoint.fn, endpoint.obj, *msg.args)
                                else:
                                    d = self._as_future(endpoint.fn, endpoint.obj)
                        else:
                            if msg.kwargs:
                                if msg.args:
                                    d = self._as_future(endpoint.fn, *msg.args, **msg.kwargs)
                                else:
                                    d = self._as_future(endpoint.fn, **msg.kwargs)
                            else:
                                if msg.args:
                                    d = self._as_future(endpoint.fn, *msg.args)
                                else:
                                    d = self._as_future(endpoint.fn)

                        def success(res):
                            del self._invocations[msg.request]

                            if isinstance(res, types.CallResult):
                                reply = message.Yield(msg.request, args=res.results, kwargs=res.kwresults)
                            else:
                                reply = message.Yield(msg.request, args=[res])
                            self._transport.send(reply)

                        def error(err):
                            if self.traceback_app:
                                # if asked to marshal the traceback within the WAMP error message, extract it
                                # noinspection PyCallingNonCallable
                                tb = StringIO()
                                err.printTraceback(file=tb)
                                tb = tb.getvalue().splitlines()
                            else:
                                tb = None

                            if self.debug_app:
                                print("Failure while invoking procedure {0} registered under '{1}' ({2}):".format(endpoint.fn, endpoint.procedure, msg.registration))
                                print(err)

                            del self._invocations[msg.request]

                            if hasattr(err, 'value'):
                                exc = err.value
                            else:
                                exc = err
                            reply = self._message_from_exception(message.Invocation.MESSAGE_TYPE, msg.request, exc, tb)
                            self._transport.send(reply)

                        self._invocations[msg.request] = d

                        self._add_future_callbacks(d, success, error)

            elif isinstance(msg, message.Interrupt):

                if msg.request not in self._invocations:
                    raise ProtocolError("INTERRUPT received for non-pending invocation {0}".format(msg.request))
                else:
                    # noinspection PyBroadException
                    try:
                        self._invocations[msg.request].cancel()
                    except Exception:
                        if self.debug:
                            print("could not cancel call {0}".format(msg.request))
                    finally:
                        del self._invocations[msg.request]

            elif isinstance(msg, message.Registered):

                if msg.request in self._register_reqs:
                    d, obj, fn, procedure, options = self._register_reqs.pop(msg.request)
                    self._registrations[msg.registration] = Endpoint(obj, fn, procedure, options)
                    r = Registration(self, msg.registration)
                    self._resolve_future(d, r)
                else:
                    raise ProtocolError("REGISTERED received for non-pending request ID {0}".format(msg.request))

            elif isinstance(msg, message.Unregistered):

                if msg.request in self._unregister_reqs:
                    d, registration = self._unregister_reqs.pop(msg.request)
                    if registration.id in self._registrations:
                        del self._registrations[registration.id]
                    registration.active = False
                    self._resolve_future(d, None)
                else:
                    raise ProtocolError("UNREGISTERED received for non-pending request ID {0}".format(msg.request))

            elif isinstance(msg, message.Error):

                d = None

                # ERROR reply to PUBLISH
                if msg.request_type == message.Publish.MESSAGE_TYPE and msg.request in self._publish_reqs:
                    d = self._publish_reqs.pop(msg.request)[0]

                # ERROR reply to SUBSCRIBE
                elif msg.request_type == message.Subscribe.MESSAGE_TYPE and msg.request in self._subscribe_reqs:
                    d = self._subscribe_reqs.pop(msg.request)[0]

                # ERROR reply to UNSUBSCRIBE
                elif msg.request_type == message.Unsubscribe.MESSAGE_TYPE and msg.request in self._unsubscribe_reqs:
                    d = self._unsubscribe_reqs.pop(msg.request)[0]

                # ERROR reply to REGISTER
                elif msg.request_type == message.Register.MESSAGE_TYPE and msg.request in self._register_reqs:
                    d = self._register_reqs.pop(msg.request)[0]

                # ERROR reply to UNREGISTER
                elif msg.request_type == message.Unregister.MESSAGE_TYPE and msg.request in self._unregister_reqs:
                    d = self._unregister_reqs.pop(msg.request)[0]

                # ERROR reply to CALL
                elif msg.request_type == message.Call.MESSAGE_TYPE and msg.request in self._call_reqs:
                    d = self._call_reqs.pop(msg.request)[0]

                if d:
                    self._reject_future(d, self._exception_from_message(msg))
                else:
                    raise ProtocolError("WampAppSession.onMessage(): ERROR received for non-pending request_type {0} and request ID {1}".format(msg.request_type, msg.request))

            elif isinstance(msg, message.Heartbeat):

                pass  # FIXME

            else:

                raise ProtocolError("Unexpected message {0}".format(msg.__class__))

    # noinspection PyUnusedLocal
    def onClose(self, wasClean):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onClose`
        """
        self._transport = None

        if self._session_id:

            # fire callback and close the transport
            try:
                self.onLeave(types.CloseDetails())
            except Exception as e:
                if self.debug:
                    print("exception raised in onLeave callback: {0}".format(e))

            self._session_id = None

        self.onDisconnect()

    def onChallenge(self, challenge):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onChallenge`
        """
        raise Exception("received authentication challenge, but onChallenge not implemented")

    def onJoin(self, details):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onJoin`
        """

    def onLeave(self, details):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onLeave`
        """
        self.disconnect()

    def leave(self, reason=None, log_message=None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.leave`
        """
        if not self._session_id:
            raise Exception("not joined")

        if not self._goodbye_sent:
            if not reason:
                reason = u"wamp.close.normal"
            msg = wamp.message.Goodbye(reason=reason, message=log_message)
            self._transport.send(msg)
            self._goodbye_sent = True
        else:
            raise SessionNotReady(u"Already requested to close the session")

    def publish(self, topic, *args, **kwargs):
        """
        Implements :func:`autobahn.wamp.interfaces.IPublisher.publish`
        """
        if six.PY2 and type(topic) == str:
            topic = six.u(topic)
        assert(type(topic) == six.text_type)

        if not self._transport:
            raise exception.TransportLost()

        request = util.id()

        if 'options' in kwargs and isinstance(kwargs['options'], types.PublishOptions):
            opts = kwargs.pop('options')
            msg = message.Publish(request, topic, args=args, kwargs=kwargs, **opts.options)
        else:
            opts = None
            msg = message.Publish(request, topic, args=args, kwargs=kwargs)

        if opts and opts.options['acknowledge'] is True:
            d = self._create_future()
            self._publish_reqs[request] = d, opts
            self._transport.send(msg)
            return d
        else:
            self._transport.send(msg)
            return

    def subscribe(self, handler, topic=None, options=None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISubscriber.subscribe`
        """
        assert((callable(handler) and topic is not None) or hasattr(handler, '__class__'))
        if topic and six.PY2 and type(topic) == str:
            topic = six.u(topic)
        assert(topic is None or type(topic) == six.text_type)
        assert(options is None or isinstance(options, types.SubscribeOptions))

        if not self._transport:
            raise exception.TransportLost()

        def _subscribe(obj, handler, topic, options):
            request = util.id()

            d = self._create_future()
            self._subscribe_reqs[request] = (d, obj, handler, topic, options)

            if options is not None:
                msg = message.Subscribe(request, topic, **options.options)
            else:
                msg = message.Subscribe(request, topic)

            self._transport.send(msg)
            return d

        if callable(handler):

            # subscribe a single handler
            return _subscribe(None, handler, topic, options)

        else:

            # subscribe all methods on an object decorated with "wamp.subscribe"
            dl = []
            for k in inspect.getmembers(handler.__class__, is_method_or_function):
                proc = k[1]
                if "_wampuris" in proc.__dict__:
                    pat = proc.__dict__["_wampuris"][0]
                    if pat.is_handler():
                        uri = pat.uri()
                        subopts = options or pat.subscribe_options()
                        dl.append(_subscribe(handler, proc, uri, subopts))
            return self._gather_futures(dl, consume_exceptions=True)

    def _unsubscribe(self, subscription):
        """
        Called from :meth:`autobahn.wamp.protocol.Subscription.unsubscribe`
        """
        assert(isinstance(subscription, Subscription))
        assert subscription.active
        assert(subscription.id in self._subscriptions)

        if not self._transport:
            raise exception.TransportLost()

        request = util.id()

        d = self._create_future()
        self._unsubscribe_reqs[request] = (d, subscription)

        msg = message.Unsubscribe(request, subscription.id)

        self._transport.send(msg)
        return d

    def call(self, procedure, *args, **kwargs):
        """
        Implements :func:`autobahn.wamp.interfaces.ICaller.call`
        """
        if six.PY2 and type(procedure) == str:
            procedure = six.u(procedure)
        assert(isinstance(procedure, six.text_type))

        if not self._transport:
            raise exception.TransportLost()

        request = util.id()

        if 'options' in kwargs and isinstance(kwargs['options'], types.CallOptions):
            opts = kwargs.pop('options')
            msg = message.Call(request, procedure, args=args, kwargs=kwargs, **opts.options)
        else:
            opts = None
            msg = message.Call(request, procedure, args=args, kwargs=kwargs)

        # FIXME
        # def canceller(_d):
        #   cancel_msg = message.Cancel(request)
        #   self._transport.send(cancel_msg)
        # d = Deferred(canceller)
        d = self._create_future()
        self._call_reqs[request] = d, opts

        self._transport.send(msg)
        return d

    def register(self, endpoint, procedure=None, options=None):
        """
        Implements :func:`autobahn.wamp.interfaces.ICallee.register`
        """
        assert((callable(endpoint) and procedure is not None) or hasattr(endpoint, '__class__'))
        if procedure and six.PY2 and type(procedure) == str:
            procedure = six.u(procedure)
        assert(procedure is None or type(procedure) == six.text_type)
        assert(options is None or isinstance(options, types.RegisterOptions))

        if not self._transport:
            raise exception.TransportLost()

        def _register(obj, endpoint, procedure, options):
            request = util.id()

            d = self._create_future()
            self._register_reqs[request] = (d, obj, endpoint, procedure, options)

            if options is not None:
                msg = message.Register(request, procedure, **options.options)
            else:
                msg = message.Register(request, procedure)

            self._transport.send(msg)
            return d

        if callable(endpoint):
            # register a single callable
            return _register(None, endpoint, procedure, options)

        else:
            # register all methods on an object
            # decorated with "wamp.register"
            dl = []

            for k in inspect.getmembers(endpoint.__class__, is_method_or_function):
                proc = k[1]
                if "_wampuris" in proc.__dict__:
                    pat = proc.__dict__["_wampuris"][0]
                    if pat.is_endpoint():
                        uri = pat.uri()
                        dl.append(_register(endpoint, proc, uri, options))
            return self._gather_futures(dl, consume_exceptions=True)

    def _unregister(self, registration):
        """
        Called from :meth:`autobahn.wamp.protocol.Registration.unregister`
        """
        assert(isinstance(registration, Registration))
        assert registration.active
        assert(registration.id in self._registrations)

        if not self._transport:
            raise exception.TransportLost()

        request = util.id()

        d = self._create_future()
        self._unregister_reqs[request] = (d, registration)

        msg = message.Unregister(request, registration.id)

        self._transport.send(msg)
        return d


IPublisher.register(ApplicationSession)
ISubscriber.register(ApplicationSession)
ICaller.register(ApplicationSession)
# ICallee.register(ApplicationSession)  # FIXME: ".register" collides with the ABC "register" method
ITransportHandler.register(ApplicationSession)


class ApplicationSessionFactory:
    """
    WAMP endpoint session factory.
    """

    session = ApplicationSession
    """
   WAMP application session class to be used in this factory.
   """

    def __init__(self, config=None):
        """

        :param config: The default component configuration.
        :type config: instance of :class:`autobahn.wamp.types.ComponentConfig`
        """
        self.config = config or types.ComponentConfig(realm=u"default")

    def __call__(self):
        """
        Creates a new WAMP application session.

        :returns: -- An instance of the WAMP application session class as
                     given by `self.session`.
        """
        session = self.session(self.config)
        session.factory = self
        return session



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import json

from autobahn import util
from autobahn.wamp.exception import ProtocolError

__all__ = ('RoleFeatures',
           'RoleBrokerFeatures',
           'RoleSubscriberFeatures',
           'RolePublisherFeatures',
           'RoleDealerFeatures',
           'RoleCallerFeatures',
           'RoleCalleeFeatures',
           'ROLE_NAME_TO_CLASS')


class RoleFeatures(util.EqualityMixin):

    ROLE = None

    def __str__(self):
        return json.dumps(self.__dict__)

    def __repr__(self):
        configured_options = {}
        for k, v in self.__dict__.items():
            if v is not None:
                configured_options[k] = v
        return "{0}({1})".format(self.ROLE, ", ".join([k + '=' + str(v)
                                                       for k, v in configured_options.items()]))

    def _check_all_bool(self):
        # check feature attributes
        for k in self.__dict__:
            if not k.startswith('_') and k != 'ROLE':
                if getattr(self, k) is not None and type(getattr(self, k)) != bool:
                    raise ProtocolError("invalid type {0} for feature '{1}' for role '{2}'".format(getattr(self, k), k, self.ROLE))


class RoleCommonPubSubFeatures(RoleFeatures):

    def __init__(self,
                 publisher_identification=None,
                 partitioned_pubsub=None):

        self.publisher_identification = publisher_identification
        self.partitioned_pubsub = partitioned_pubsub


class RoleBrokerFeatures(RoleCommonPubSubFeatures):

    ROLE = u'broker'

    def __init__(self,
                 subscriber_blackwhite_listing=None,
                 publisher_exclusion=None,
                 publication_trustlevels=None,
                 pattern_based_subscription=None,
                 subscriber_metaevents=None,
                 subscriber_list=None,
                 event_history=None,
                 **kwargs):
        self.subscriber_blackwhite_listing = subscriber_blackwhite_listing
        self.publisher_exclusion = publisher_exclusion
        self.publication_trustlevels = publication_trustlevels
        self.pattern_based_subscription = pattern_based_subscription
        self.subscriber_metaevents = subscriber_metaevents
        self.subscriber_list = subscriber_list
        self.event_history = event_history
        RoleCommonPubSubFeatures.__init__(self, **kwargs)
        self._check_all_bool()


class RoleSubscriberFeatures(RoleCommonPubSubFeatures):

    ROLE = u'subscriber'

    def __init__(self,
                 publication_trustlevels=None,
                 pattern_based_subscription=None,
                 subscriber_metaevents=None,
                 subscriber_list=None,
                 event_history=None,
                 **kwargs):
        self.publication_trustlevels = publication_trustlevels
        self.pattern_based_subscription = pattern_based_subscription
        self.subscriber_metaevents = subscriber_metaevents
        self.subscriber_list = subscriber_list
        self.event_history = event_history
        RoleCommonPubSubFeatures.__init__(self, **kwargs)
        self._check_all_bool()


class RolePublisherFeatures(RoleCommonPubSubFeatures):

    ROLE = u'publisher'

    def __init__(self,
                 subscriber_blackwhite_listing=None,
                 publisher_exclusion=None,
                 **kwargs):
        self.subscriber_blackwhite_listing = subscriber_blackwhite_listing
        self.publisher_exclusion = publisher_exclusion
        RoleCommonPubSubFeatures.__init__(self, **kwargs)
        self._check_all_bool()


class RoleCommonRpcFeatures(RoleFeatures):

    def __init__(self,
                 caller_identification=None,
                 partitioned_rpc=None,
                 call_timeout=None,
                 call_canceling=None,
                 progressive_call_results=None):
        self.caller_identification = caller_identification
        self.partitioned_rpc = partitioned_rpc
        self.call_timeout = call_timeout
        self.call_canceling = call_canceling
        self.progressive_call_results = progressive_call_results


class RoleDealerFeatures(RoleCommonRpcFeatures):

    ROLE = u'dealer'

    def __init__(self,
                 callee_blackwhite_listing=None,
                 caller_exclusion=None,
                 call_trustlevels=None,
                 pattern_based_registration=None,
                 **kwargs):
        self.callee_blackwhite_listing = callee_blackwhite_listing
        self.caller_exclusion = caller_exclusion
        self.call_trustlevels = call_trustlevels
        self.pattern_based_registration = pattern_based_registration
        RoleCommonRpcFeatures.__init__(self, **kwargs)
        self._check_all_bool()


class RoleCallerFeatures(RoleCommonRpcFeatures):

    ROLE = u'caller'

    def __init__(self,
                 callee_blackwhite_listing=None,
                 caller_exclusion=None,
                 **kwargs):
        self.callee_blackwhite_listing = callee_blackwhite_listing
        self.caller_exclusion = caller_exclusion
        RoleCommonRpcFeatures.__init__(self, **kwargs)
        self._check_all_bool()


class RoleCalleeFeatures(RoleCommonRpcFeatures):

    ROLE = u'callee'

    def __init__(self,
                 call_trustlevels=None,
                 pattern_based_registration=None,
                 **kwargs):
        self.call_trustlevels = call_trustlevels
        self.pattern_based_registration = pattern_based_registration
        RoleCommonRpcFeatures.__init__(self, **kwargs)
        self._check_all_bool()


ROLE_NAME_TO_CLASS = {
    u'broker': RoleBrokerFeatures,
    u'subscriber': RoleSubscriberFeatures,
    u'publisher': RolePublisherFeatures,
    u'dealer': RoleDealerFeatures,
    u'caller': RoleCallerFeatures,
    u'callee': RoleCalleeFeatures,
}



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import abc
import six

__all__ = ('IWebSocketChannel',
           'IWebSocketChannelFrameApi',
           'IWebSocketChannelStreamingApi')


@six.add_metaclass(abc.ABCMeta)
class IWebSocketChannel(object):
    """
    A WebSocket channel is a bidirectional, full-duplex, ordered, reliable message channel
    over a WebSocket connection as specified in RFC6455.

    This interface defines a message-based API to WebSocket plus auxiliary hooks
    and methods.
    """

    @abc.abstractmethod
    def onConnect(self, requestOrResponse):
        """
        Callback fired during WebSocket opening handshake when a client connects (to a server with
        request from client) or when server connection established (by a client with response from
        server).

        :param requestOrResponse: Connection request (for servers) or response (for clients).
        :type requestOrResponse: Instance of :class:`autobahn.websocket.protocol.ConnectionRequest`
           or :class:`autobahn.websocket.protocol.ConnectionResponse`.

        :returns:
           When this callback is fired on a WebSocket server, you may return one of the
           following:

           1. ``None``: Connection accepted (no subprotocol)
           2. ``str``: Connection accepted with given subprotocol
           3. ``(subprotocol, headers)``: Connection accepted with given ``subprotocol`` (which
              also may be ``None``) and set the given HTTP ``headers`` (e.g. cookies). ``headers``
              must be a ``dict`` with ``str`` keys and values for the HTTP header values to set.

              If a given header value is a non-string iterable (e.g. list or tuple), a separate
              header line will be sent for each item in the iterable.

           If the client announced one or multiple subprotocols, the server MUST select
           one of the given list.
        """

    @abc.abstractmethod
    def onOpen(self):
        """
        Callback fired when the initial WebSocket opening handshake was completed.
        You now can send and receive WebSocket messages.
        """

    @abc.abstractmethod
    def sendMessage(self, payload, isBinary=False, fragmentSize=None, sync=False, doNotCompress=False):
        """
        Send a WebSocket message over the connection to the peer.

        :param payload: The message payload.
        :type payload: bytes
        :param isBinary: ``True`` when payload is binary, else the payload must be UTF-8 encoded text.
        :type isBinary: bool
        :param fragmentSize: Fragment message into WebSocket fragments of this size (the last frame
           potentially being shorter).
        :type fragmentSize: int
        :param sync: If ``True``, try to force data onto the wire immediately.

           .. warning::
              Do NOT use this feature for normal applications.
              Performance likely will suffer significantly.
              This feature is mainly here for use by Autobahn|Testsuite.
        :type sync: bool
        :param doNotCompress: Iff ``True``, never compress this message. This only applies to
                              Hybi-Mode and only when WebSocket compression has been negotiated on
                              the WebSocket connection. Use when you know the payload
                              incompressible (e.g. encrypted or already compressed).
        :type doNotCompress: bool
        """

    @abc.abstractmethod
    def onMessage(self, payload, isBinary):
        """
        Callback fired when a complete WebSocket message was received.

        :param payload: Message payload (UTF-8 encoded text or binary). Can also be empty when
           the WebSocket message contained no payload.
        :type payload: bytes
        :param isBinary: ``True`` iff payload is binary, else the payload is UTF-8 encoded text.
        :type isBinary: bool
        """

    @abc.abstractmethod
    def sendClose(self, code=None, reason=None):
        """
        Starts a WebSocket closing handshake tearing down the WebSocket connection.

        :param code: An optional close status code (``1000`` for normal close or ``3000-4999`` for
           application specific close).
        :type code: int
        :param reason: An optional close reason (a string that when present, a status
           code MUST also be present).
        :type reason: str
        """

    @abc.abstractmethod
    def onClose(self, wasClean, code, reason):
        """
        Callback fired when the WebSocket connection has been closed (WebSocket closing
        handshake has been finished or the connection was closed uncleanly).

        :param wasClean: ``True`` iff the WebSocket connection was closed cleanly.
        :type wasClean: bool
        :param code: Close status code (as sent by the WebSocket peer).
        :type code: int or None
        :param reason: Close reason (as sent by the WebSocket peer).
        :type reason: unicode or None
        """

    @abc.abstractmethod
    def sendPreparedMessage(self, preparedMsg):
        """
        Send a message that was previously prepared with :func:`autobahn.websocket.protocol.WebSocketFactory.prepareMessage`.

        :param prepareMessage: A previously prepared message.
        :type prepareMessage: Instance of :class:`autobahn.websocket.protocol.PreparedMessage`.
        """

    @abc.abstractmethod
    def sendPing(self, payload=None):
        """
        Send a WebSocket ping to the peer.

        A peer is expected to pong back the payload a soon as "practical". When more than
        one ping is outstanding at a peer, the peer may elect to respond only to the last ping.

        :param payload: An (optional) arbitrary payload of length **less than 126** octets.
        :type payload: bytes or None
        """

    @abc.abstractmethod
    def onPing(self, payload):
        """
        Callback fired when a WebSocket ping was received. A default implementation responds
        by sending a WebSocket pong.

        :param payload: Payload of ping (when there was any). Can be arbitrary, up to `125` octets.
        :type payload: bytes
        """

    @abc.abstractmethod
    def sendPong(self, payload=None):
        """
        Send a WebSocket pong to the peer.

        A WebSocket pong may be sent unsolicited. This serves as a unidirectional heartbeat.
        A response to an unsolicited pong is "not expected".

        :param payload: An (optional) arbitrary payload of length < 126 octets.
        :type payload: bytes
        """

    @abc.abstractmethod
    def onPong(self, payload):
        """
        Callback fired when a WebSocket pong was received. A default implementation does nothing.

        :param payload: Payload of pong (when there was any). Can be arbitrary, up to 125 octets.
        :type payload: bytes
        """


class IWebSocketChannelFrameApi(IWebSocketChannel):
    """
    Frame-based API to a WebSocket channel.
    """

    @abc.abstractmethod
    def onMessageBegin(self, isBinary):
        """
        Callback fired when receiving of a new WebSocket message has begun.

        :param isBinary: ``True`` if payload is binary, else the payload is UTF-8 encoded text.
        :type isBinary: bool
        """

    @abc.abstractmethod
    def onMessageFrame(self, payload):
        """
        Callback fired when a complete WebSocket message frame for a previously begun
        WebSocket message has been received.

        :param payload: Message frame payload (a list of chunks received).
        :type payload: list of bytes
        """

    @abc.abstractmethod
    def onMessageEnd(self):
        """
        Callback fired when a WebSocket message has been completely received (the last
        WebSocket frame for that message has been received).
        """

    @abc.abstractmethod
    def beginMessage(self, isBinary=False, doNotCompress=False):
        """
        Begin sending a new WebSocket message.

        :param isBinary: ``True`` if payload is binary, else the payload must be UTF-8 encoded text.
        :type isBinary: bool
        :param doNotCompress: If ``True``, never compress this message. This only applies to
           Hybi-Mode and only when WebSocket compression has been negotiated on the WebSocket
           connection. Use when you know the payload incompressible (e.g. encrypted or
           already compressed).
        :type doNotCompress: bool
        """

    @abc.abstractmethod
    def sendMessageFrame(self, payload, sync=False):
        """
        When a message has been previously begun, send a complete message frame in one go.

        :param payload: The message frame payload. When sending a text message, the payload must
                        be UTF-8 encoded already.
        :type payload: bytes
        :param sync: If ``True``, try to force data onto the wire immediately.

           .. warning::
              Do NOT use this feature for normal applications.
              Performance likely will suffer significantly.
              This feature is mainly here for use by Autobahn|Testsuite.
        :type sync: bool
        """

    @abc.abstractmethod
    def endMessage(self):
        """
        End a message previously begun message. No more frames may be sent (for that message).
        You have to begin a new message before sending again.
        """


class IWebSocketChannelStreamingApi(IWebSocketChannelFrameApi):
    """
    Streaming API to a WebSocket channel.
    """

    @abc.abstractmethod
    def onMessageFrameBegin(self, length):
        """
        Callback fired when receiving a new message frame has begun.
        A default implementation will prepare to buffer message frame data.

        :param length: Payload length of message frame which is subsequently received.
        :type length: int
        """

    @abc.abstractmethod
    def onMessageFrameData(self, payload):
        """
        Callback fired when receiving data within a previously begun message frame.
        A default implementation will buffer data for frame.

        :param payload: Partial payload for message frame.
        :type payload: bytes
        """

    @abc.abstractmethod
    def onMessageFrameEnd(self):
        """
        Callback fired when a previously begun message frame has been completely received.
        A default implementation will flatten the buffered frame data and
        fire `onMessageFrame`.
        """

    @abc.abstractmethod
    def beginMessageFrame(self, length):
        """
        Begin sending a new message frame.

        :param length: Length of the frame which is to be started. Must be less or equal **2^63**.
        :type length: int
        """

    @abc.abstractmethod
    def sendMessageFrameData(self, payload, sync=False):
        """
        Send out data when within a message frame (message was begun, frame was begun).
        Note that the frame is automatically ended when enough data has been sent.
        In other words, there is no ``endMessageFrame``, since you have begun the frame
        specifying the frame length, which implicitly defined the frame end. This is different
        from messages, which you begin *and* end explicitly , since a message can contain
        an unlimited number of frames.

        :param payload: Frame payload to send.
        :type payload: bytes
        :param sync: If ``True``, try to force data onto the wire immediately.

           .. warning::
              Do NOT use this feature for normal applications.
              Performance likely will suffer significantly.
              This feature is mainly here for use by Autobahn|Testsuite.
        :type sync: bool

        :returns: When the currently sent message frame is still incomplete, returns octets
           remaining to be sent. When the frame is complete, returns **0**. Otherwise the amount
           of unconsumed data in payload argument is returned.
        :rtype: int
        """



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import six


# use Cython implementation of XorMasker validator if available
##
try:
    from wsaccel.xormask import XorMaskerNull, createXorMasker

except ImportError:
    # fallback to pure Python implementation

    # http://stackoverflow.com/questions/15014310/python3-xrange-lack-hurts
    try:
        xrange
    except NameError:
        # Python 3
        # noinspection PyShadowingBuiltins
        xrange = range

    from array import array

    class XorMaskerNull:

        # noinspection PyUnusedLocal
        def __init__(self, mask=None):
            self.ptr = 0

        def pointer(self):
            return self.ptr

        def reset(self):
            self.ptr = 0

        def process(self, data):
            self.ptr += len(data)
            return data

    class XorMaskerSimple:

        def __init__(self, mask):
            assert len(mask) == 4
            self.ptr = 0
            self.msk = array('B', mask)

        def pointer(self):
            return self.ptr

        def reset(self):
            self.ptr = 0

        def process(self, data):
            dlen = len(data)
            payload = array('B', data)
            for k in xrange(dlen):
                payload[k] ^= self.msk[self.ptr & 3]
                self.ptr += 1
            if six.PY3:
                return payload.tobytes()
            else:
                return payload.tostring()

    class XorMaskerShifted1:

        def __init__(self, mask):
            assert len(mask) == 4
            self.ptr = 0
            self.mskarray = [array('B'), array('B'), array('B'), array('B')]
            if six.PY3:
                for j in xrange(4):
                    self.mskarray[0].append(mask[j & 3])
                    self.mskarray[1].append(mask[(j + 1) & 3])
                    self.mskarray[2].append(mask[(j + 2) & 3])
                    self.mskarray[3].append(mask[(j + 3) & 3])
            else:
                for j in xrange(4):
                    self.mskarray[0].append(ord(mask[j & 3]))
                    self.mskarray[1].append(ord(mask[(j + 1) & 3]))
                    self.mskarray[2].append(ord(mask[(j + 2) & 3]))
                    self.mskarray[3].append(ord(mask[(j + 3) & 3]))

        def pointer(self):
            return self.ptr

        def reset(self):
            self.ptr = 0

        def process(self, data):
            dlen = len(data)
            payload = array('B', data)
            msk = self.mskarray[self.ptr & 3]
            for k in xrange(dlen):
                payload[k] ^= msk[k & 3]
            self.ptr += dlen
            if six.PY3:
                return payload.tobytes()
            else:
                return payload.tostring()

    def createXorMasker(mask, length=None):
        if length is None or length < 128:
            return XorMaskerSimple(mask)
        else:
            return XorMaskerShifted1(mask)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import absolute_import

import re
import six

import autobahn
from autobahn import util
from autobahn.wamp.exception import ProtocolError
from autobahn.wamp.interfaces import IMessage
from autobahn.wamp.role import ROLE_NAME_TO_CLASS

__all__ = ('Message',
           'Hello',
           'Welcome',
           'Abort',
           'Challenge',
           'Authenticate',
           'Goodbye',
           'Heartbeat',
           'Error',
           'Publish',
           'Published',
           'Subscribe',
           'Subscribed',
           'Unsubscribe',
           'Unsubscribed',
           'Event',
           'Call',
           'Cancel',
           'Result',
           'Register',
           'Registered',
           'Unregister',
           'Unregistered',
           'Invocation',
           'Interrupt',
           'Yield',
           'check_or_raise_uri',
           'check_or_raise_id')


# strict URI check allowing empty URI components
_URI_PAT_STRICT_EMPTY = re.compile(r"^(([0-9a-z_]+\.)|\.)*([0-9a-z_]+)?$")

# loose URI check allowing empty URI components
_URI_PAT_LOOSE_EMPTY = re.compile(r"^(([^\s\.#]+\.)|\.)*([^\s\.#]+)?$")

# strict URI check disallowing empty URI components
_URI_PAT_STRICT_NON_EMPTY = re.compile(r"^([0-9a-z_]+\.)*([0-9a-z_]+)$")

# loose URI check disallowing empty URI components
_URI_PAT_LOOSE_NON_EMPTY = re.compile(r"^([^\s\.#]+\.)*([^\s\.#]+)$")


def check_or_raise_uri(value, message=u"WAMP message invalid", strict=False, allowEmptyComponents=False):
    """
    Check a value for being a valid WAMP URI.

    If the value is not a valid WAMP URI is invalid, raises :class:`autobahn.wamp.exception.ProtocolError`.
    Otherwise return the value.

    :param value: The value to check.
    :type value: unicode
    :param message: Prefix for message in exception raised when value is invalid.
    :type message: unicode
    :param strict: If ``True``, do a strict check on the URI (the WAMP spec SHOULD behavior).
    :type strict: bool
    :param allowEmptyComponents: If ``True``, allow empty URI components (for pattern based
       subscriptions and registrations).
    :type allowEmptyComponents: bool

    :returns: The URI value (if valid).
    :rtype: unicode
    :raises: instance of :class:`autobahn.wamp.exception.ProtocolError`
    """
    if type(value) != six.text_type:
        raise ProtocolError(u"{0}: invalid type {1} for URI".format(message, type(value)))

    if strict:
        if allowEmptyComponents:
            pat = _URI_PAT_STRICT_EMPTY
        else:
            pat = _URI_PAT_STRICT_NON_EMPTY
    else:
        if allowEmptyComponents:
            pat = _URI_PAT_LOOSE_EMPTY
        else:
            pat = _URI_PAT_LOOSE_NON_EMPTY

    if not pat.match(value):
        raise ProtocolError(u"{0}: invalid value '{1}' for URI".format(message, value))
    else:
        return value


def check_or_raise_id(value, message=u"WAMP message invalid"):
    """
    Check a value for being a valid WAMP ID.

    If the value is not a valid WAMP ID, raises :class:`autobahn.wamp.exception.ProtocolError`.
    Otherwise return the value.

    :param value: The value to check.
    :type value: int
    :param message: Prefix for message in exception raised when value is invalid.
    :type message: unicode

    :returns: The ID value (if valid).
    :rtype: int
    :raises: instance of :class:`autobahn.wamp.exception.ProtocolError`
    """
    if type(value) not in six.integer_types:
        raise ProtocolError(u"{0}: invalid type {1} for ID".format(message, type(value)))
    if value < 0 or value > 9007199254740992:  # 2**53
        raise ProtocolError(u"{0}: invalid value {1} for ID".format(message, value))
    return value


def check_or_raise_extra(value, message=u"WAMP message invalid"):
    """
    Check a value for being a valid WAMP extra dictionary.

    If the value is not a valid WAMP extra dictionary, raises :class:`autobahn.wamp.exception.ProtocolError`.
    Otherwise return the value.

    :param value: The value to check.
    :type value: dict
    :param message: Prefix for message in exception raised when value is invalid.
    :type message: unicode

    :returns: The extra dictionary (if valid).
    :rtype: dict
    :raises: instance of :class:`autobahn.wamp.exception.ProtocolError`
    """
    if type(value) != dict:
        raise ProtocolError(u"{0}: invalid type {1}".format(message, type(value)))
    for k in value.keys():
        if type(k) != six.text_type:
            raise ProtocolError(u"{0}: invalid type {1} for key '{2}'".format(message, type(k), k))
    return value


class Message(util.EqualityMixin):
    """
    WAMP message base class. Implements :class:`autobahn.wamp.interfaces.IMessage`.

    .. note:: This is not supposed to be instantiated.
    """

    def __init__(self):
        # serialization cache: mapping from ISerializer instances to serialized bytes
        self._serialized = {}

    def uncache(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.uncache`
        """
        self._serialized = {}

    def serialize(self, serializer):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.serialize`
        """
        # only serialize if not cached ..
        if serializer not in self._serialized:
            self._serialized[serializer] = serializer.serialize(self.marshal())
        return self._serialized[serializer]


IMessage.register(Message)


class Hello(Message):
    """
    A WAMP ``HELLO`` message.

    Format: ``[HELLO, Realm|uri, Details|dict]``
    """

    MESSAGE_TYPE = 1
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, realm, roles, authmethods=None, authid=None):
        """

        :param realm: The URI of the WAMP realm to join.
        :type realm: unicode
        :param roles: The WAMP roles to announce.
        :type roles: list of :class:`autobahn.wamp.role.RoleFeatures`
        :param authmethods: The authentication methods to announce.
        :type authmethods: list of unicode or None
        :param authid: The authentication ID to announce.
        :type authid: unicode or None
        """
        assert(type(realm) == six.text_type)
        assert(type(roles) == list)
        for role in roles:
            assert(isinstance(role, autobahn.wamp.role.RoleFeatures))
        if authmethods:
            assert(type(authmethods) == list)
            for authmethod in authmethods:
                assert(type(authmethod) == six.text_type)
        assert(authid is None or type(authid) == six.text_type)

        Message.__init__(self)
        self.realm = realm
        self.roles = roles
        self.authmethods = authmethods
        self.authid = authid

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Hello.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for HELLO".format(len(wmsg)))

        realm = check_or_raise_uri(wmsg[1], u"'realm' in HELLO")
        details = check_or_raise_extra(wmsg[2], u"'details' in HELLO")

        roles = []

        if u'roles' not in details:
            raise ProtocolError(u"missing mandatory roles attribute in options in HELLO")

        details_roles = check_or_raise_extra(details[u'roles'], u"'roles' in 'details' in HELLO")

        if len(details_roles) == 0:
            raise ProtocolError(u"empty 'roles' in 'details' in HELLO")

        for role in details_roles:
            if role not in ROLE_NAME_TO_CLASS:
                raise ProtocolError("invalid role '{0}' in 'roles' in 'details' in HELLO".format(role))

            role_cls = ROLE_NAME_TO_CLASS[role]

            details_role = check_or_raise_extra(details_roles[role], "role '{0}' in 'roles' in 'details' in HELLO".format(role))

            if u'features' in details_role:
                check_or_raise_extra(details_role[u'features'], "'features' in role '{0}' in 'roles' in 'details' in HELLO".format(role))

                # FIXME: skip unknown attributes
                role_features = role_cls(**details_role[u'features'])

            else:
                role_features = role_cls()

            roles.append(role_features)

        authmethods = None
        if u'authmethods' in details:
            details_authmethods = details[u'authmethods']
            if type(details_authmethods) != list:
                raise ProtocolError("invalid type {0} for 'authmethods' detail in HELLO".format(type(details_authmethods)))

            for auth_method in details_authmethods:
                if type(auth_method) != six.text_type:
                    raise ProtocolError("invalid type {0} for item in 'authmethods' detail in HELLO".format(type(auth_method)))

            authmethods = details_authmethods

        authid = None
        if u'authid' in details:
            details_authid = details[u'authid']
            if type(details_authid) != six.text_type:
                raise ProtocolError("invalid type {0} for 'authid' detail in HELLO".format(type(details_authid)))

            authid = details_authid

        obj = Hello(realm, roles, authmethods, authid)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {u'roles': {}}
        for role in self.roles:
            details[u'roles'][role.ROLE] = {}
            for feature in role.__dict__:
                if not feature.startswith('_') and feature != 'ROLE' and getattr(role, feature) is not None:
                    if u'features' not in details[u'roles'][role.ROLE]:
                        details[u'roles'][role.ROLE] = {u'features': {}}
                    details[u'roles'][role.ROLE][u'features'][six.u(feature)] = getattr(role, feature)

        if self.authmethods:
            details[u'authmethods'] = self.authmethods

        if self.authid:
            details[u'authid'] = self.authid

        return [Hello.MESSAGE_TYPE, self.realm, details]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP HELLO Message (realm = {0}, roles = {1}, authmethods = {2}, authid = {3})".format(self.realm, self.roles, self.authmethods, self.authid)


class Welcome(Message):
    """
    A WAMP ``WELCOME`` message.

    Format: ``[WELCOME, Session|id, Details|dict]``
    """

    MESSAGE_TYPE = 2
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, session, roles, authid=None, authrole=None, authmethod=None, authprovider=None):
        """

        :param session: The WAMP session ID the other peer is assigned.
        :type session: int
        :param roles: The WAMP roles to announce.
        :type roles: list of :class:`autobahn.wamp.role.RoleFeatures`
        :param authid: The authentication ID assigned.
        :type authid: unicode or None
        :param authrole: The authentication role assigned.
        :type authrole: unicode or None
        :param authmethod: The authentication method in use.
        :type authmethod: unicode or None
        :param authprovider: The authentication method in use.
        :type authprovider: unicode or None
        """
        assert(type(session) in six.integer_types)
        assert(type(roles) == list)
        for role in roles:
            assert(isinstance(role, autobahn.wamp.role.RoleFeatures))
        assert(authid is None or type(authid) == six.text_type)
        assert(authrole is None or type(authrole) == six.text_type)
        assert(authmethod is None or type(authmethod) == six.text_type)
        assert(authprovider is None or type(authprovider) == six.text_type)

        Message.__init__(self)
        self.session = session
        self.roles = roles
        self.authid = authid
        self.authrole = authrole
        self.authmethod = authmethod
        self.authprovider = authprovider

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Welcome.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for WELCOME".format(len(wmsg)))

        session = check_or_raise_id(wmsg[1], u"'session' in WELCOME")
        details = check_or_raise_extra(wmsg[2], u"'details' in WELCOME")

        authid = details.get(u'authid', None)
        authrole = details.get(u'authrole', None)
        authmethod = details.get(u'authmethod', None)
        authprovider = details.get(u'authprovider', None)

        roles = []

        if u'roles' not in details:
            raise ProtocolError(u"missing mandatory roles attribute in options in WELCOME")

        details_roles = check_or_raise_extra(details['roles'], u"'roles' in 'details' in WELCOME")

        if len(details_roles) == 0:
            raise ProtocolError(u"empty 'roles' in 'details' in WELCOME")

        for role in details_roles:
            if role not in ROLE_NAME_TO_CLASS:
                raise ProtocolError("invalid role '{0}' in 'roles' in 'details' in WELCOME".format(role))

            role_cls = ROLE_NAME_TO_CLASS[role]

            if u'features' in details_roles[role]:
                check_or_raise_extra(details_roles[role][u'features'], "'features' in role '{0}' in 'roles' in 'details' in WELCOME".format(role))

                # FIXME: skip unknown attributes
                role_features = role_cls(**details_roles[role][u'features'])

            else:
                role_features = role_cls()

            roles.append(role_features)

        obj = Welcome(session, roles, authid, authrole, authmethod, authprovider)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {
            u'roles': {}
        }

        if self.authid:
            details[u'authid'] = self.authid

        if self.authrole:
            details[u'authrole'] = self.authrole

        if self.authrole:
            details[u'authmethod'] = self.authmethod

        if self.authprovider:
            details[u'authprovider'] = self.authprovider

        for role in self.roles:
            details[u'roles'][role.ROLE] = {}
            for feature in role.__dict__:
                if not feature.startswith('_') and feature != 'ROLE' and getattr(role, feature) is not None:
                    if u'features' not in details[u'roles'][role.ROLE]:
                        details[u'roles'][role.ROLE] = {u'features': {}}
                    details[u'roles'][role.ROLE][u'features'][six.u(feature)] = getattr(role, feature)

        return [Welcome.MESSAGE_TYPE, self.session, details]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP WELCOME Message (session = {0}, roles = {1}, authid = {2}, authrole = {3}, authmethod = {4}, authprovider = {5})".format(self.session, self.roles, self.authid, self.authrole, self.authmethod, self.authprovider)


class Abort(Message):
    """
    A WAMP ``ABORT`` message.

    Format: ``[ABORT, Details|dict, Reason|uri]``
    """

    MESSAGE_TYPE = 3
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, reason, message=None):
        """

        :param reason: WAMP or application error URI for aborting reason.
        :type reason: unicode
        :param message: Optional human-readable closing message, e.g. for logging purposes.
        :type message: unicode or None
        """
        assert(type(reason) == six.text_type)
        assert(message is None or type(message) == six.text_type)

        Message.__init__(self)
        self.reason = reason
        self.message = message

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Abort.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for ABORT".format(len(wmsg)))

        details = check_or_raise_extra(wmsg[1], u"'details' in ABORT")
        reason = check_or_raise_uri(wmsg[2], u"'reason' in ABORT")

        message = None

        if u'message' in details:

            details_message = details[u'message']
            if type(details_message) != six.text_type:
                raise ProtocolError("invalid type {0} for 'message' detail in ABORT".format(type(details_message)))

            message = details_message

        obj = Abort(reason, message)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {}
        if self.message:
            details[u'message'] = self.message

        return [Abort.MESSAGE_TYPE, details, self.reason]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP ABORT Message (message = {0}, reason = {1})".format(self.message, self.reason)


class Challenge(Message):
    """
    A WAMP ``CHALLENGE`` message.

    Format: ``[CHALLENGE, Method|string, Extra|dict]``
    """

    MESSAGE_TYPE = 4
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, method, extra=None):
        """

        :param method: The authentication method.
        :type method: unicode
        :param extra: Authentication method specific information.
        :type extra: dict or None
        """
        assert(type(method) == six.text_type)
        assert(extra is None or type(extra) == dict)

        Message.__init__(self)
        self.method = method
        self.extra = extra or {}

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Challenge.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for CHALLENGE".format(len(wmsg)))

        method = wmsg[1]
        if type(method) != six.text_type:
            raise ProtocolError("invalid type {0} for 'method' in CHALLENGE".format(type(method)))

        extra = check_or_raise_extra(wmsg[2], u"'extra' in CHALLENGE")

        obj = Challenge(method, extra)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Challenge.MESSAGE_TYPE, self.method, self.extra]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP CHALLENGE Message (method = {0}, extra = {1})".format(self.method, self.extra)


class Authenticate(Message):
    """
    A WAMP ``AUTHENTICATE`` message.

    Format: ``[AUTHENTICATE, Signature|string, Extra|dict]``
    """

    MESSAGE_TYPE = 5
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, signature, extra=None):
        """

        :param signature: The signature for the authentication challenge.
        :type signature: unicode
        :param extra: Authentication method specific information.
        :type extra: dict or None
        """
        assert(type(signature) == six.text_type)
        assert(extra is None or type(extra) == dict)

        Message.__init__(self)
        self.signature = signature
        self.extra = extra or {}

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Authenticate.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for AUTHENTICATE".format(len(wmsg)))

        signature = wmsg[1]
        if type(signature) != six.text_type:
            raise ProtocolError("invalid type {0} for 'signature' in AUTHENTICATE".format(type(signature)))

        extra = check_or_raise_extra(wmsg[2], u"'extra' in AUTHENTICATE")

        obj = Authenticate(signature, extra)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Authenticate.MESSAGE_TYPE, self.signature, self.extra]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP AUTHENTICATE Message (signature = {0}, extra = {1})".format(self.signature, self.extra)


class Goodbye(Message):
    """
    A WAMP ``GOODBYE`` message.

    Format: ``[GOODBYE, Details|dict, Reason|uri]``
    """

    MESSAGE_TYPE = 6
    """
   The WAMP message code for this type of message.
   """

    DEFAULT_REASON = u"wamp.goodbye.normal"
    """
   Default WAMP closing reason.
   """

    def __init__(self, reason=DEFAULT_REASON, message=None):
        """

        :param reason: Optional WAMP or application error URI for closing reason.
        :type reason: unicode
        :param message: Optional human-readable closing message, e.g. for logging purposes.
        :type message: unicode or None
        """
        assert(type(reason) == six.text_type)
        assert(message is None or type(message) == six.text_type)

        Message.__init__(self)
        self.reason = reason
        self.message = message

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Goodbye.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for GOODBYE".format(len(wmsg)))

        details = check_or_raise_extra(wmsg[1], u"'details' in GOODBYE")
        reason = check_or_raise_uri(wmsg[2], u"'reason' in GOODBYE")

        message = None

        if u'message' in details:

            details_message = details[u'message']
            if type(details_message) != six.text_type:
                raise ProtocolError("invalid type {0} for 'message' detail in GOODBYE".format(type(details_message)))

            message = details_message

        obj = Goodbye(reason, message)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {}
        if self.message:
            details[u'message'] = self.message

        return [Goodbye.MESSAGE_TYPE, details, self.reason]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP GOODBYE Message (message = {0}, reason = {1})".format(self.message, self.reason)


class Heartbeat(Message):
    """
    A WAMP ``HEARTBEAT`` message.

    Formats:

    * ``[HEARTBEAT, Incoming|integer, Outgoing|integer]``
    * ``[HEARTBEAT, Incoming|integer, Outgoing|integer, Discard|string]``
    """

    MESSAGE_TYPE = 7
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, incoming, outgoing, discard=None):
        """

        :param incoming: Last incoming heartbeat processed from peer.
        :type incoming: int
        :param outgoing: Outgoing heartbeat.
        :type outgoing: int
        :param discard: Optional data that is discarded by peer.
        :type discard: unicode or None
        """
        assert(type(incoming) in six.integer_types)
        assert(type(outgoing) in six.integer_types)
        assert(discard is None or type(discard) == six.text_type)

        Message.__init__(self)
        self.incoming = incoming
        self.outgoing = outgoing
        self.discard = discard

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Heartbeat.MESSAGE_TYPE)

        if len(wmsg) not in [3, 4]:
            raise ProtocolError("invalid message length {0} for HEARTBEAT".format(len(wmsg)))

        incoming = wmsg[1]

        if type(incoming) not in six.integer_types:
            raise ProtocolError("invalid type {0} for 'incoming' in HEARTBEAT".format(type(incoming)))

        if incoming < 0:  # must be non-negative
            raise ProtocolError("invalid value {0} for 'incoming' in HEARTBEAT".format(incoming))

        outgoing = wmsg[2]

        if type(outgoing) not in six.integer_types:
            raise ProtocolError("invalid type {0} for 'outgoing' in HEARTBEAT".format(type(outgoing)))

        if outgoing <= 0:  # must be positive
            raise ProtocolError("invalid value {0} for 'outgoing' in HEARTBEAT".format(outgoing))

        discard = None
        if len(wmsg) > 3:
            discard = wmsg[3]
            if type(discard) != six.text_type:
                raise ProtocolError("invalid type {0} for 'discard' in HEARTBEAT".format(type(discard)))

        obj = Heartbeat(incoming, outgoing, discard=discard)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        if self.discard:
            return [Heartbeat.MESSAGE_TYPE, self.incoming, self.outgoing, self.discard]
        else:
            return [Heartbeat.MESSAGE_TYPE, self.incoming, self.outgoing]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP HEARTBEAT Message (incoming {0}, outgoing = {1}, len(discard) = {2})".format(self.incoming, self.outgoing, len(self.discard) if self.discard else None)


class Error(Message):
    """
    A WAMP ``ERROR`` message.

    Formats:

    * ``[ERROR, REQUEST.Type|int, REQUEST.Request|id, Details|dict, Error|uri]``
    * ``[ERROR, REQUEST.Type|int, REQUEST.Request|id, Details|dict, Error|uri, Arguments|list]``
    * ``[ERROR, REQUEST.Type|int, REQUEST.Request|id, Details|dict, Error|uri, Arguments|list, ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 8
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request_type, request, error, args=None, kwargs=None):
        """

        :param request_type: The WAMP message type code for the original request.
        :type request_type: int
        :param request: The WAMP request ID of the original request (`Call`, `Subscribe`, ...) this error occurred for.
        :type request: int
        :param error: The WAMP or application error URI for the error that occurred.
        :type error: unicode
        :param args: Positional values for application-defined exception.
           Must be serializable using any serializers in use.
        :type args: list or None
        :param kwargs: Keyword values for application-defined exception.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        """
        assert(type(request_type) in six.integer_types)
        assert(type(request) in six.integer_types)
        assert(type(error) == six.text_type)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)

        Message.__init__(self)
        self.request_type = request_type
        self.request = request
        self.error = error
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Error.MESSAGE_TYPE)

        if len(wmsg) not in (5, 6, 7):
            raise ProtocolError("invalid message length {0} for ERROR".format(len(wmsg)))

        request_type = wmsg[1]
        if type(request_type) not in six.integer_types:
            raise ProtocolError("invalid type {0} for 'request_type' in ERROR".format(request_type))

        if request_type not in [Subscribe.MESSAGE_TYPE,
                                Unsubscribe.MESSAGE_TYPE,
                                Publish.MESSAGE_TYPE,
                                Register.MESSAGE_TYPE,
                                Unregister.MESSAGE_TYPE,
                                Call.MESSAGE_TYPE,
                                Invocation.MESSAGE_TYPE]:
            raise ProtocolError("invalid value {0} for 'request_type' in ERROR".format(request_type))

        request = check_or_raise_id(wmsg[2], u"'request' in ERROR")
        check_or_raise_extra(wmsg[3], u"'details' in ERROR")
        error = check_or_raise_uri(wmsg[4], u"'error' in ERROR")

        args = None
        if len(wmsg) > 5:
            args = wmsg[5]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in ERROR".format(type(args)))

        kwargs = None
        if len(wmsg) > 6:
            kwargs = wmsg[6]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in ERROR".format(type(kwargs)))

        obj = Error(request_type, request, error, args=args, kwargs=kwargs)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {}

        if self.kwargs:
            return [self.MESSAGE_TYPE, self.request_type, self.request, details, self.error, self.args, self.kwargs]
        elif self.args:
            return [self.MESSAGE_TYPE, self.request_type, self.request, details, self.error, self.args]
        else:
            return [self.MESSAGE_TYPE, self.request_type, self.request, details, self.error]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP Error Message (request_type = {0}, request = {1}, error = {2}, args = {3}, kwargs = {4})".format(self.request_type, self.request, self.error, self.args, self.kwargs)


class Publish(Message):
    """
    A WAMP ``PUBLISH`` message.

    Formats:

    * ``[PUBLISH, Request|id, Options|dict, Topic|uri]``
    * ``[PUBLISH, Request|id, Options|dict, Topic|uri, Arguments|list]``
    * ``[PUBLISH, Request|id, Options|dict, Topic|uri, Arguments|list, ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 16
    """
   The WAMP message code for this type of message.
   """

    def __init__(self,
                 request,
                 topic,
                 args=None,
                 kwargs=None,
                 acknowledge=None,
                 excludeMe=None,
                 exclude=None,
                 eligible=None,
                 discloseMe=None):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param topic: The WAMP or application URI of the PubSub topic the event should
           be published to.
        :type topic: unicode
        :param args: Positional values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type args: list or tuple or None
        :param kwargs: Keyword values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        :param acknowledge: If True, acknowledge the publication with a success or
           error response.
        :type acknowledge: bool or None
        :param excludeMe: If ``True``, exclude the publisher from receiving the event, even
           if he is subscribed (and eligible).
        :type excludeMe: bool or None
        :param exclude: List of WAMP session IDs to exclude from receiving this event.
        :type exclude: list of int or None
        :param eligible: List of WAMP session IDs eligible to receive this event.
        :type eligible: list of int or None
        :param discloseMe: If True, request to disclose the publisher of this event
           to subscribers.
        :type discloseMe: bool or None
        """
        assert(type(request) in six.integer_types)
        assert(type(topic) == six.text_type)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)
        assert(acknowledge is None or type(acknowledge) == bool)
        assert(excludeMe is None or type(excludeMe) == bool)
        assert(exclude is None or type(exclude) == list)
        assert(eligible is None or type(eligible) == list)
        assert(discloseMe is None or type(discloseMe) == bool)

        Message.__init__(self)
        self.request = request
        self.topic = topic
        self.args = args
        self.kwargs = kwargs
        self.acknowledge = acknowledge
        self.excludeMe = excludeMe
        self.exclude = exclude
        self.eligible = eligible
        self.discloseMe = discloseMe

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Publish.MESSAGE_TYPE)

        if len(wmsg) not in (4, 5, 6):
            raise ProtocolError("invalid message length {0} for PUBLISH".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in PUBLISH")
        options = check_or_raise_extra(wmsg[2], u"'options' in PUBLISH")
        topic = check_or_raise_uri(wmsg[3], u"'topic' in PUBLISH")

        args = None
        if len(wmsg) > 4:
            args = wmsg[4]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in PUBLISH".format(type(args)))

        kwargs = None
        if len(wmsg) > 5:
            kwargs = wmsg[5]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in PUBLISH".format(type(kwargs)))

        acknowledge = None
        excludeMe = None
        exclude = None
        eligible = None
        discloseMe = None

        if u'acknowledge' in options:

            option_acknowledge = options[u'acknowledge']
            if type(option_acknowledge) != bool:
                raise ProtocolError("invalid type {0} for 'acknowledge' option in PUBLISH".format(type(option_acknowledge)))

            acknowledge = option_acknowledge

        if u'exclude_me' in options:

            option_excludeMe = options[u'exclude_me']
            if type(option_excludeMe) != bool:
                raise ProtocolError("invalid type {0} for 'exclude_me' option in PUBLISH".format(type(option_excludeMe)))

            excludeMe = option_excludeMe

        if u'exclude' in options:

            option_exclude = options[u'exclude']
            if type(option_exclude) != list:
                raise ProtocolError("invalid type {0} for 'exclude' option in PUBLISH".format(type(option_exclude)))

            for sessionId in option_exclude:
                if type(sessionId) not in six.integer_types:
                    raise ProtocolError("invalid type {0} for value in 'exclude' option in PUBLISH".format(type(sessionId)))

            exclude = option_exclude

        if u'eligible' in options:

            option_eligible = options[u'eligible']
            if type(option_eligible) != list:
                raise ProtocolError("invalid type {0} for 'eligible' option in PUBLISH".format(type(option_eligible)))

            for sessionId in option_eligible:
                if type(sessionId) not in six.integer_types:
                    raise ProtocolError("invalid type {0} for value in 'eligible' option in PUBLISH".format(type(sessionId)))

            eligible = option_eligible

        if u'disclose_me' in options:

            option_discloseMe = options[u'disclose_me']
            if type(option_discloseMe) != bool:
                raise ProtocolError("invalid type {0} for 'disclose_me' option in PUBLISH".format(type(option_discloseMe)))

            discloseMe = option_discloseMe

        obj = Publish(request,
                      topic,
                      args=args,
                      kwargs=kwargs,
                      acknowledge=acknowledge,
                      excludeMe=excludeMe,
                      exclude=exclude,
                      eligible=eligible,
                      discloseMe=discloseMe)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.acknowledge is not None:
            options[u'acknowledge'] = self.acknowledge
        if self.excludeMe is not None:
            options[u'exclude_me'] = self.excludeMe
        if self.exclude is not None:
            options[u'exclude'] = self.exclude
        if self.eligible is not None:
            options[u'eligible'] = self.eligible
        if self.discloseMe is not None:
            options[u'disclose_me'] = self.discloseMe

        if self.kwargs:
            return [Publish.MESSAGE_TYPE, self.request, options, self.topic, self.args, self.kwargs]
        elif self.args:
            return [Publish.MESSAGE_TYPE, self.request, options, self.topic, self.args]
        else:
            return [Publish.MESSAGE_TYPE, self.request, options, self.topic]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP PUBLISH Message (request = {0}, topic = {1}, args = {2}, kwargs = {3}, acknowledge = {4}, excludeMe = {5}, exclude = {6}, eligible = {7}, discloseMe = {8})".format(self.request, self.topic, self.args, self.kwargs, self.acknowledge, self.excludeMe, self.exclude, self.eligible, self.discloseMe)


class Published(Message):
    """
    A WAMP ``PUBLISHED`` message.

    Format: ``[PUBLISHED, PUBLISH.Request|id, Publication|id]``
    """

    MESSAGE_TYPE = 17
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, publication):
        """

        :param request: The request ID of the original `PUBLISH` request.
        :type request: int
        :param publication: The publication ID for the published event.
        :type publication: int
        """
        assert(type(request) in six.integer_types)
        assert(type(publication) in six.integer_types)

        Message.__init__(self)
        self.request = request
        self.publication = publication

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Published.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for PUBLISHED".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in PUBLISHED")
        publication = check_or_raise_id(wmsg[2], u"'publication' in PUBLISHED")

        obj = Published(request, publication)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Published.MESSAGE_TYPE, self.request, self.publication]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP PUBLISHED Message (request = {0}, publication = {1})".format(self.request, self.publication)


class Subscribe(Message):
    """
    A WAMP ``SUBSCRIBE`` message.

    Format: ``[SUBSCRIBE, Request|id, Options|dict, Topic|uri]``
    """

    MESSAGE_TYPE = 32
    """
   The WAMP message code for this type of message.
   """

    MATCH_EXACT = u'exact'
    MATCH_PREFIX = u'prefix'
    MATCH_WILDCARD = u'wildcard'

    def __init__(self, request, topic, match=MATCH_EXACT):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param topic: The WAMP or application URI of the PubSub topic to subscribe to.
        :type topic: unicode
        :param match: The topic matching method to be used for the subscription.
        :type match: unicode
        """
        assert(type(request) in six.integer_types)
        assert(type(topic) == six.text_type)
        assert(match is None or type(match) == six.text_type)
        assert(match is None or match in [self.MATCH_EXACT, self.MATCH_PREFIX, self.MATCH_WILDCARD])

        Message.__init__(self)
        self.request = request
        self.topic = topic
        self.match = match

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Subscribe.MESSAGE_TYPE)

        if len(wmsg) != 4:
            raise ProtocolError("invalid message length {0} for SUBSCRIBE".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in SUBSCRIBE")
        options = check_or_raise_extra(wmsg[2], u"'options' in SUBSCRIBE")
        topic = check_or_raise_uri(wmsg[3], u"'topic' in SUBSCRIBE", allowEmptyComponents=True)

        match = Subscribe.MATCH_EXACT

        if u'match' in options:

            option_match = options[u'match']
            if type(option_match) != six.text_type:
                raise ProtocolError("invalid type {0} for 'match' option in SUBSCRIBE".format(type(option_match)))

            if option_match not in [Subscribe.MATCH_EXACT, Subscribe.MATCH_PREFIX, Subscribe.MATCH_WILDCARD]:
                raise ProtocolError("invalid value {0} for 'match' option in SUBSCRIBE".format(option_match))

            match = option_match

        obj = Subscribe(request, topic, match)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.match and self.match != Subscribe.MATCH_EXACT:
            options[u'match'] = self.match

        return [Subscribe.MESSAGE_TYPE, self.request, options, self.topic]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP SUBSCRIBE Message (request = {0}, topic = {1}, match = {2})".format(self.request, self.topic, self.match)


class Subscribed(Message):
    """
    A WAMP ``SUBSCRIBED`` message.

    Format: ``[SUBSCRIBED, SUBSCRIBE.Request|id, Subscription|id]``
    """

    MESSAGE_TYPE = 33
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, subscription):
        """

        :param request: The request ID of the original ``SUBSCRIBE`` request.
        :type request: int
        :param subscription: The subscription ID for the subscribed topic (or topic pattern).
        :type subscription: int
        """
        assert(type(request) in six.integer_types)
        assert(type(subscription) in six.integer_types)

        Message.__init__(self)
        self.request = request
        self.subscription = subscription

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Subscribed.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for SUBSCRIBED".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in SUBSCRIBED")
        subscription = check_or_raise_id(wmsg[2], u"'subscription' in SUBSCRIBED")

        obj = Subscribed(request, subscription)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Subscribed.MESSAGE_TYPE, self.request, self.subscription]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP SUBSCRIBED Message (request = {0}, subscription = {1})".format(self.request, self.subscription)


class Unsubscribe(Message):
    """
    A WAMP ``UNSUBSCRIBE`` message.

    Format: ``[UNSUBSCRIBE, Request|id, SUBSCRIBED.Subscription|id]``
    """

    MESSAGE_TYPE = 34
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, subscription):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param subscription: The subscription ID for the subscription to unsubscribe from.
        :type subscription: int
        """
        assert(type(request) in six.integer_types)
        assert(type(subscription) in six.integer_types)

        Message.__init__(self)
        self.request = request
        self.subscription = subscription

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Unsubscribe.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for WAMP UNSUBSCRIBE".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in UNSUBSCRIBE")
        subscription = check_or_raise_id(wmsg[2], u"'subscription' in UNSUBSCRIBE")

        obj = Unsubscribe(request, subscription)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Unsubscribe.MESSAGE_TYPE, self.request, self.subscription]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP UNSUBSCRIBE Message (request = {0}, subscription = {1})".format(self.request, self.subscription)


class Unsubscribed(Message):
    """
    A WAMP ``UNSUBSCRIBED`` message.

    Format: ``[UNSUBSCRIBED, UNSUBSCRIBE.Request|id]``
    """

    MESSAGE_TYPE = 35
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request):
        """

        :param request: The request ID of the original ``UNSUBSCRIBE`` request.
        :type request: int
        """
        assert(type(request) in six.integer_types)

        Message.__init__(self)
        self.request = request

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Unsubscribed.MESSAGE_TYPE)

        if len(wmsg) != 2:
            raise ProtocolError("invalid message length {0} for UNSUBSCRIBED".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in UNSUBSCRIBED")

        obj = Unsubscribed(request)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Unsubscribed.MESSAGE_TYPE, self.request]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP UNSUBSCRIBED Message (request = {0})".format(self.request)


class Event(Message):
    """
    A WAMP ``EVENT`` message.

    Formats:

    * ``[EVENT, SUBSCRIBED.Subscription|id, PUBLISHED.Publication|id, Details|dict]``
    * ``[EVENT, SUBSCRIBED.Subscription|id, PUBLISHED.Publication|id, Details|dict, PUBLISH.Arguments|list]``
    * ``[EVENT, SUBSCRIBED.Subscription|id, PUBLISHED.Publication|id, Details|dict, PUBLISH.Arguments|list, PUBLISH.ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 36
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, subscription, publication, args=None, kwargs=None, publisher=None):
        """

        :param subscription: The subscription ID this event is dispatched under.
        :type subscription: int
        :param publication: The publication ID of the dispatched event.
        :type publication: int
        :param args: Positional values for application-defined exception.
           Must be serializable using any serializers in use.
        :type args: list or tuple or None
        :param kwargs: Keyword values for application-defined exception.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        :param publisher: If present, the WAMP session ID of the publisher of this event.
        :type publisher: int or None
        """
        assert(type(subscription) in six.integer_types)
        assert(type(publication) in six.integer_types)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)
        assert(publisher is None or type(publisher) in six.integer_types)

        Message.__init__(self)
        self.subscription = subscription
        self.publication = publication
        self.args = args
        self.kwargs = kwargs
        self.publisher = publisher

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Event.MESSAGE_TYPE)

        if len(wmsg) not in (4, 5, 6):
            raise ProtocolError("invalid message length {0} for EVENT".format(len(wmsg)))

        subscription = check_or_raise_id(wmsg[1], u"'subscription' in EVENT")
        publication = check_or_raise_id(wmsg[2], u"'publication' in EVENT")
        details = check_or_raise_extra(wmsg[3], u"'details' in EVENT")

        args = None
        if len(wmsg) > 4:
            args = wmsg[4]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in EVENT".format(type(args)))

        kwargs = None
        if len(wmsg) > 5:
            kwargs = wmsg[5]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in EVENT".format(type(kwargs)))

        publisher = None
        if u'publisher' in details:

            detail_publisher = details[u'publisher']
            if type(detail_publisher) not in six.integer_types:
                raise ProtocolError("invalid type {0} for 'publisher' detail in EVENT".format(type(detail_publisher)))

            publisher = detail_publisher

        obj = Event(subscription,
                    publication,
                    args=args,
                    kwargs=kwargs,
                    publisher=publisher)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {}

        if self.publisher is not None:
            details[u'publisher'] = self.publisher

        if self.kwargs:
            return [Event.MESSAGE_TYPE, self.subscription, self.publication, details, self.args, self.kwargs]
        elif self.args:
            return [Event.MESSAGE_TYPE, self.subscription, self.publication, details, self.args]
        else:
            return [Event.MESSAGE_TYPE, self.subscription, self.publication, details]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP EVENT Message (subscription = {0}, publication = {1}, args = {2}, kwargs = {3}, publisher = {4})".format(self.subscription, self.publication, self.args, self.kwargs, self.publisher)


class Call(Message):
    """
    A WAMP ``CALL`` message.

    Formats:

    * ``[CALL, Request|id, Options|dict, Procedure|uri]``
    * ``[CALL, Request|id, Options|dict, Procedure|uri, Arguments|list]``
    * ``[CALL, Request|id, Options|dict, Procedure|uri, Arguments|list, ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 48
    """
   The WAMP message code for this type of message.
   """

    def __init__(self,
                 request,
                 procedure,
                 args=None,
                 kwargs=None,
                 timeout=None,
                 receive_progress=None,
                 discloseMe=None):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param procedure: The WAMP or application URI of the procedure which should be called.
        :type procedure: unicode
        :param args: Positional values for application-defined call arguments.
           Must be serializable using any serializers in use.
        :type args: list or tuple or None
        :param kwargs: Keyword values for application-defined call arguments.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        :param timeout: If present, let the callee automatically cancel
           the call after this ms.
        :type timeout: int or None
        :param receive_progress: If ``True``, indicates that the caller wants to receive
           progressive call results.
        :type receive_progress: bool or None
        :param discloseMe: If ``True``, the caller requests to disclose itself to the callee.
        :type discloseMe: bool or None
        """
        assert(type(request) in six.integer_types)
        assert(type(procedure) == six.text_type)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)
        assert(timeout is None or type(timeout) in six.integer_types)
        assert(receive_progress is None or type(receive_progress) == bool)
        assert(discloseMe is None or type(discloseMe) == bool)

        Message.__init__(self)
        self.request = request
        self.procedure = procedure
        self.args = args
        self.kwargs = kwargs
        self.timeout = timeout
        self.receive_progress = receive_progress
        self.discloseMe = discloseMe

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Call.MESSAGE_TYPE)

        if len(wmsg) not in (4, 5, 6):
            raise ProtocolError("invalid message length {0} for CALL".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in CALL")
        options = check_or_raise_extra(wmsg[2], u"'options' in CALL")
        procedure = check_or_raise_uri(wmsg[3], u"'procedure' in CALL")

        args = None
        if len(wmsg) > 4:
            args = wmsg[4]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in CALL".format(type(args)))

        kwargs = None
        if len(wmsg) > 5:
            kwargs = wmsg[5]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in CALL".format(type(kwargs)))

        timeout = None
        if u'timeout' in options:

            option_timeout = options[u'timeout']
            if type(option_timeout) not in six.integer_types:
                raise ProtocolError("invalid type {0} for 'timeout' option in CALL".format(type(option_timeout)))

            if option_timeout < 0:
                raise ProtocolError("invalid value {0} for 'timeout' option in CALL".format(option_timeout))

            timeout = option_timeout

        receive_progress = None
        if u'receive_progress' in options:

            option_receive_progress = options[u'receive_progress']
            if type(option_receive_progress) != bool:
                raise ProtocolError("invalid type {0} for 'receive_progress' option in CALL".format(type(option_receive_progress)))

            receive_progress = option_receive_progress

        discloseMe = None
        if u'disclose_me' in options:

            option_discloseMe = options[u'disclose_me']
            if type(option_discloseMe) != bool:
                raise ProtocolError("invalid type {0} for 'disclose_me' option in CALL".format(type(option_discloseMe)))

            discloseMe = option_discloseMe

        obj = Call(request,
                   procedure,
                   args=args,
                   kwargs=kwargs,
                   timeout=timeout,
                   receive_progress=receive_progress,
                   discloseMe=discloseMe)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.timeout is not None:
            options[u'timeout'] = self.timeout

        if self.receive_progress is not None:
            options[u'receive_progress'] = self.receive_progress

        if self.discloseMe is not None:
            options[u'disclose_me'] = self.discloseMe

        if self.kwargs:
            return [Call.MESSAGE_TYPE, self.request, options, self.procedure, self.args, self.kwargs]
        elif self.args:
            return [Call.MESSAGE_TYPE, self.request, options, self.procedure, self.args]
        else:
            return [Call.MESSAGE_TYPE, self.request, options, self.procedure]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP CALL Message (request = {0}, procedure = {1}, args = {2}, kwargs = {3}, timeout = {4}, receive_progress = {5}, discloseMe = {6})".format(self.request, self.procedure, self.args, self.kwargs, self.timeout, self.receive_progress, self.discloseMe)


class Cancel(Message):
    """
    A WAMP ``CANCEL`` message.

    Format: ``[CANCEL, CALL.Request|id, Options|dict]``
    """

    MESSAGE_TYPE = 49
    """
   The WAMP message code for this type of message.
   """

    SKIP = u'skip'
    ABORT = u'abort'
    KILL = u'kill'

    def __init__(self, request, mode=None):
        """

        :param request: The WAMP request ID of the original `CALL` to cancel.
        :type request: int
        :param mode: Specifies how to cancel the call (``"skip"``, ``"abort"`` or ``"kill"``).
        :type mode: unicode or None
        """
        assert(type(request) in six.integer_types)
        assert(mode is None or type(mode) == six.text_type)
        assert(mode in [None, self.SKIP, self.ABORT, self.KILL])

        Message.__init__(self)
        self.request = request
        self.mode = mode

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Cancel.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for CANCEL".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in CANCEL")
        options = check_or_raise_extra(wmsg[2], u"'options' in CANCEL")

        # options
        ##
        mode = None

        if u'mode' in options:

            option_mode = options[u'mode']
            if type(option_mode) != six.text_type:
                raise ProtocolError("invalid type {0} for 'mode' option in CANCEL".format(type(option_mode)))

            if option_mode not in [Cancel.SKIP, Cancel.ABORT, Cancel.KILL]:
                raise ProtocolError("invalid value '{0}' for 'mode' option in CANCEL".format(option_mode))

            mode = option_mode

        obj = Cancel(request, mode=mode)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.mode is not None:
            options[u'mode'] = self.mode

        return [Cancel.MESSAGE_TYPE, self.request, options]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP CANCEL Message (request = {0}, mode = '{1}'')".format(self.request, self.mode)


class Result(Message):
    """
    A WAMP ``RESULT`` message.

    Formats:

    * ``[RESULT, CALL.Request|id, Details|dict]``
    * ``[RESULT, CALL.Request|id, Details|dict, YIELD.Arguments|list]``
    * ``[RESULT, CALL.Request|id, Details|dict, YIELD.Arguments|list, YIELD.ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 50
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, args=None, kwargs=None, progress=None):
        """

        :param request: The request ID of the original `CALL` request.
        :type request: int
        :param args: Positional values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type args: list or tuple or None
        :param kwargs: Keyword values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        :param progress: If ``True``, this result is a progressive call result, and subsequent
           results (or a final error) will follow.
        :type progress: bool or None
        """
        assert(type(request) in six.integer_types)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)
        assert(progress is None or type(progress) == bool)

        Message.__init__(self)
        self.request = request
        self.args = args
        self.kwargs = kwargs
        self.progress = progress

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Result.MESSAGE_TYPE)

        if len(wmsg) not in (3, 4, 5):
            raise ProtocolError("invalid message length {0} for RESULT".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in RESULT")
        details = check_or_raise_extra(wmsg[2], u"'details' in RESULT")

        args = None
        if len(wmsg) > 3:
            args = wmsg[3]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in RESULT".format(type(args)))

        kwargs = None
        if len(wmsg) > 4:
            kwargs = wmsg[4]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in RESULT".format(type(kwargs)))

        progress = None

        if u'progress' in details:

            detail_progress = details[u'progress']
            if type(detail_progress) != bool:
                raise ProtocolError("invalid type {0} for 'progress' option in RESULT".format(type(detail_progress)))

            progress = detail_progress

        obj = Result(request, args=args, kwargs=kwargs, progress=progress)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        details = {}

        if self.progress is not None:
            details[u'progress'] = self.progress

        if self.kwargs:
            return [Result.MESSAGE_TYPE, self.request, details, self.args, self.kwargs]
        elif self.args:
            return [Result.MESSAGE_TYPE, self.request, details, self.args]
        else:
            return [Result.MESSAGE_TYPE, self.request, details]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP RESULT Message (request = {0}, args = {1}, kwargs = {2}, progress = {3})".format(self.request, self.args, self.kwargs, self.progress)


class Register(Message):
    """
    A WAMP ``REGISTER`` message.

    Format: ``[REGISTER, Request|id, Options|dict, Procedure|uri]``
    """

    MESSAGE_TYPE = 64
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, procedure, pkeys=None, discloseCaller=None, discloseCallerTransport=None):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param procedure: The WAMP or application URI of the RPC endpoint provided.
        :type procedure: unicode
        :param pkeys: The endpoint can work for this list of application partition keys.
        :type pkeys: list of int or None
        :param discloseCaller: If ``True``, the (registering) callee requests to disclose
           the identity of callers whenever called.
        :type discloseCaller: bool or None
        :param discloseCallerTransport: If ``True``, the (registering) calle requests to disclose
           transport level information of caller whenever called. This only takes effect if ``discloseCaller == True``.
        :type discloseCallerTransport: bool
        """
        assert(type(request) in six.integer_types)
        assert(type(procedure) == six.text_type)
        assert(pkeys is None or type(pkeys) == list)
        if pkeys:
            for k in pkeys:
                assert(type(k) in six.integer_types)
        assert(discloseCaller is None or type(discloseCaller) == bool)
        assert(discloseCallerTransport is None or type(discloseCallerTransport) == bool)

        Message.__init__(self)
        self.request = request
        self.procedure = procedure
        self.pkeys = pkeys
        self.discloseCaller = discloseCaller
        self.discloseCallerTransport = discloseCallerTransport

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Register.MESSAGE_TYPE)

        if len(wmsg) != 4:
            raise ProtocolError("invalid message length {0} for REGISTER".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in REGISTER")
        options = check_or_raise_extra(wmsg[2], u"'options' in REGISTER")
        procedure = check_or_raise_uri(wmsg[3], u"'procedure' in REGISTER", allowEmptyComponents=True)

        pkeys = None
        discloseCaller = None
        discloseCallerTransport = None

        if u'pkeys' in options:

            option_pkeys = options[u'pkeys']
            if type(option_pkeys) != list:
                raise ProtocolError("invalid type {0} for 'pkeys' option in REGISTER".format(type(option_pkeys)))

            for pk in option_pkeys:
                if type(pk) not in six.integer_types:
                    raise ProtocolError("invalid type for value '{0}' in 'pkeys' option in REGISTER".format(type(pk)))

            pkeys = option_pkeys

        if u'disclose_caller' in options:

            option_discloseCaller = options[u'disclose_caller']
            if type(option_discloseCaller) != bool:
                raise ProtocolError("invalid type {0} for 'disclose_caller' option in REGISTER".format(type(option_discloseCaller)))

            discloseCaller = option_discloseCaller

        if u'disclose_caller_transport' in options:

            option_discloseCallerTransport = options[u'disclose_caller_transport']
            if type(option_discloseCallerTransport) != bool:
                raise ProtocolError("invalid type {0} for 'disclose_caller_transport' option in REGISTER".format(type(option_discloseCallerTransport)))

            discloseCallerTransport = option_discloseCallerTransport

        obj = Register(request, procedure, pkeys=pkeys, discloseCaller=discloseCaller, discloseCallerTransport=discloseCallerTransport)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.pkeys is not None:
            options[u'pkeys'] = self.pkeys

        if self.discloseCaller is not None:
            options[u'disclose_caller'] = self.discloseCaller

        if self.discloseCallerTransport is not None:
            options[u'disclose_caller_transport'] = self.discloseCallerTransport

        return [Register.MESSAGE_TYPE, self.request, options, self.procedure]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP REGISTER Message (request = {0}, procedure = {1}, pkeys = {2}, discloseCaller = {3})".format(self.request, self.procedure, self.pkeys, self.discloseCaller)


class Registered(Message):
    """
    A WAMP ``REGISTERED`` message.

    Format: ``[REGISTERED, REGISTER.Request|id, Registration|id]``
    """

    MESSAGE_TYPE = 65
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, registration):
        """

        :param request: The request ID of the original ``REGISTER`` request.
        :type request: int
        :param registration: The registration ID for the registered procedure (or procedure pattern).
        :type registration: int
        """
        assert(type(request) in six.integer_types)
        assert(type(registration) in six.integer_types)

        Message.__init__(self)
        self.request = request
        self.registration = registration

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Registered.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for REGISTERED".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in REGISTERED")
        registration = check_or_raise_id(wmsg[2], u"'registration' in REGISTERED")

        obj = Registered(request, registration)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Registered.MESSAGE_TYPE, self.request, self.registration]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP REGISTERED Message (request = {0}, registration = {1})".format(self.request, self.registration)


class Unregister(Message):
    """
    A WAMP `UNREGISTER` message.

    Format: ``[UNREGISTER, Request|id, REGISTERED.Registration|id]``
    """

    MESSAGE_TYPE = 66
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, registration):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param registration: The registration ID for the registration to unregister.
        :type registration: int
        """
        assert(type(request) in six.integer_types)
        assert(type(registration) in six.integer_types)

        Message.__init__(self)
        self.request = request
        self.registration = registration

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Unregister.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for WAMP UNREGISTER".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in UNREGISTER")
        registration = check_or_raise_id(wmsg[2], u"'registration' in UNREGISTER")

        obj = Unregister(request, registration)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Unregister.MESSAGE_TYPE, self.request, self.registration]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP UNREGISTER Message (request = {0}, registration = {1})".format(self.request, self.registration)


class Unregistered(Message):
    """
    A WAMP ``UNREGISTERED`` message.

    Format: ``[UNREGISTERED, UNREGISTER.Request|id]``
    """

    MESSAGE_TYPE = 67
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request):
        """

        :param request: The request ID of the original ``UNREGISTER`` request.
        :type request: int
        """
        assert(type(request) in six.integer_types)

        Message.__init__(self)
        self.request = request

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Unregistered.MESSAGE_TYPE)

        if len(wmsg) != 2:
            raise ProtocolError("invalid message length {0} for UNREGISTERED".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in UNREGISTERED")

        obj = Unregistered(request)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        return [Unregistered.MESSAGE_TYPE, self.request]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP UNREGISTERED Message (request = {0})".format(self.request)


class Invocation(Message):
    """
    A WAMP ``INVOCATION`` message.

    Formats:

    * ``[INVOCATION, Request|id, REGISTERED.Registration|id, Details|dict]``
    * ``[INVOCATION, Request|id, REGISTERED.Registration|id, Details|dict, CALL.Arguments|list]``
    * ``[INVOCATION, Request|id, REGISTERED.Registration|id, Details|dict, CALL.Arguments|list, CALL.ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 68
    """
   The WAMP message code for this type of message.
   """

    def __init__(self,
                 request,
                 registration,
                 args=None,
                 kwargs=None,
                 timeout=None,
                 receive_progress=None,
                 caller=None,
                 caller_transport=None,
                 authid=None,
                 authrole=None,
                 authmethod=None):
        """

        :param request: The WAMP request ID of this request.
        :type request: int
        :param registration: The registration ID of the endpoint to be invoked.
        :type registration: int
        :param args: Positional values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type args: list or tuple or None
        :param kwargs: Keyword values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        :param timeout: If present, let the callee automatically cancels
           the invocation after this ms.
        :type timeout: int or None
        :param receive_progress: Indicates if the callee should produce progressive results.
        :type receive_progress: bool or None
        :param caller: The WAMP session ID of the caller.
        :type caller: int or None
        :param caller_transport: WAMP transport level information of the caller.
        :type caller_transport: dict or None
        :param authid: The authentication ID of the caller.
        :type authid: unicode or None
        :param authrole: The authentication role of the caller.
        :type authrole: unicode or None
        :param authmethod: The authentication method under which the caller was authenticated.
        :type authmethod: unicode or None
        """
        assert(type(request) in six.integer_types)
        assert(type(registration) in six.integer_types)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)
        assert(timeout is None or type(timeout) in six.integer_types)
        assert(receive_progress is None or type(receive_progress) == bool)
        assert(caller is None or type(caller) in six.integer_types)
        assert(caller_transport is None or type(caller_transport) == dict)
        assert(authid is None or type(authid) == six.text_type)
        assert(authrole is None or type(authrole) == six.text_type)
        assert(authmethod is None or type(authmethod) == six.text_type)

        Message.__init__(self)
        self.request = request
        self.registration = registration
        self.args = args
        self.kwargs = kwargs
        self.timeout = timeout
        self.receive_progress = receive_progress
        self.caller = caller
        self.caller_transport = caller_transport
        self.authid = authid
        self.authrole = authrole
        self.authmethod = authmethod

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Invocation.MESSAGE_TYPE)

        if len(wmsg) not in (4, 5, 6):
            raise ProtocolError("invalid message length {0} for INVOCATION".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in INVOCATION")
        registration = check_or_raise_id(wmsg[2], u"'registration' in INVOCATION")
        details = check_or_raise_extra(wmsg[3], u"'details' in INVOCATION")

        args = None
        if len(wmsg) > 4:
            args = wmsg[4]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in INVOCATION".format(type(args)))

        kwargs = None
        if len(wmsg) > 5:
            kwargs = wmsg[5]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in INVOCATION".format(type(kwargs)))

        timeout = None
        if u'timeout' in details:

            detail_timeout = details[u'timeout']
            if type(detail_timeout) not in six.integer_types:
                raise ProtocolError("invalid type {0} for 'timeout' detail in INVOCATION".format(type(detail_timeout)))

            if detail_timeout < 0:
                raise ProtocolError("invalid value {0} for 'timeout' detail in INVOCATION".format(detail_timeout))

            timeout = detail_timeout

        receive_progress = None
        if u'receive_progress' in details:

            detail_receive_progress = details[u'receive_progress']
            if type(detail_receive_progress) != bool:
                raise ProtocolError("invalid type {0} for 'receive_progress' detail in INVOCATION".format(type(detail_receive_progress)))

            receive_progress = detail_receive_progress

        caller = None
        if u'caller' in details:

            detail_caller = details[u'caller']
            if type(detail_caller) not in six.integer_types:
                raise ProtocolError("invalid type {0} for 'caller' detail in INVOCATION".format(type(detail_caller)))

            caller = detail_caller

        caller_transport = None
        if u'caller_transport' in details:

            detail_caller_transport = details[u'caller_transport']
            if type(detail_caller_transport) != dict:
                raise ProtocolError("invalid type {0} for 'caller_transport' detail in INVOCATION".format(type(caller_transport)))

            caller_transport = detail_caller_transport

        authid = None
        if u'authid' in details:

            detail_authid = details[u'authid']
            if type(detail_authid) != six.text_type:
                raise ProtocolError("invalid type {0} for 'authid' detail in INVOCATION".format(type(detail_authid)))

            authid = detail_authid

        authrole = None
        if u'authrole' in details:

            detail_authrole = details[u'authrole']
            if type(detail_authrole) != six.text_type:
                raise ProtocolError("invalid type {0} for 'authrole' detail in INVOCATION".format(type(detail_authrole)))

            authrole = detail_authrole

        authmethod = None
        if u'authmethod' in details:

            detail_authmethod = details[u'authmethod']
            if type(detail_authmethod) != six.text_type:
                raise ProtocolError("invalid type {0} for 'authmethod' detail in INVOCATION".format(type(detail_authmethod)))

            authmethod = detail_authmethod

        obj = Invocation(request,
                         registration,
                         args=args,
                         kwargs=kwargs,
                         timeout=timeout,
                         receive_progress=receive_progress,
                         caller=caller,
                         caller_transport=caller_transport,
                         authid=authid,
                         authrole=authrole,
                         authmethod=authmethod)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.timeout is not None:
            options[u'timeout'] = self.timeout

        if self.receive_progress is not None:
            options[u'receive_progress'] = self.receive_progress

        if self.caller is not None:
            options[u'caller'] = self.caller

        if self.caller_transport is not None:
            options[u'caller_transport'] = self.caller_transport

        if self.authid is not None:
            options[u'authid'] = self.authid

        if self.authrole is not None:
            options[u'authrole'] = self.authrole

        if self.authmethod is not None:
            options[u'authmethod'] = self.authmethod

        if self.kwargs:
            return [Invocation.MESSAGE_TYPE, self.request, self.registration, options, self.args, self.kwargs]
        elif self.args:
            return [Invocation.MESSAGE_TYPE, self.request, self.registration, options, self.args]
        else:
            return [Invocation.MESSAGE_TYPE, self.request, self.registration, options]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP INVOCATION Message (request = {0}, registration = {1}, args = {2}, kwargs = {3}, timeout = {4}, receive_progress = {5}, caller = {6}, caller_transport = {7}, authid = {8}, authrole = {9}, authmethod = {10})".format(self.request, self.registration, self.args, self.kwargs, self.timeout, self.receive_progress, self.caller, self.caller_transport, self.authid, self.authrole, self.authmethod)


class Interrupt(Message):
    """
    A WAMP ``INTERRUPT`` message.

    Format: ``[INTERRUPT, INVOCATION.Request|id, Options|dict]``
    """

    MESSAGE_TYPE = 69
    """
   The WAMP message code for this type of message.
   """

    ABORT = u'abort'
    KILL = u'kill'

    def __init__(self, request, mode=None):
        """

        :param request: The WAMP request ID of the original ``INVOCATION`` to interrupt.
        :type request: int
        :param mode: Specifies how to interrupt the invocation (``"abort"`` or ``"kill"``).
        :type mode: unicode or None
        """
        assert(type(request) in six.integer_types)
        assert(mode is None or type(mode) == six.text_type)
        assert(mode is None or mode in [self.ABORT, self.KILL])

        Message.__init__(self)
        self.request = request
        self.mode = mode

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Interrupt.MESSAGE_TYPE)

        if len(wmsg) != 3:
            raise ProtocolError("invalid message length {0} for INTERRUPT".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in INTERRUPT")
        options = check_or_raise_extra(wmsg[2], u"'options' in INTERRUPT")

        # options
        ##
        mode = None

        if u'mode' in options:

            option_mode = options[u'mode']
            if type(option_mode) != six.text_type:
                raise ProtocolError("invalid type {0} for 'mode' option in INTERRUPT".format(type(option_mode)))

            if option_mode not in [Interrupt.ABORT, Interrupt.KILL]:
                raise ProtocolError("invalid value '{0}' for 'mode' option in INTERRUPT".format(option_mode))

            mode = option_mode

        obj = Interrupt(request, mode=mode)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.mode is not None:
            options[u'mode'] = self.mode

        return [Interrupt.MESSAGE_TYPE, self.request, options]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP INTERRUPT Message (request = {0}, mode = '{1}')".format(self.request, self.mode)


class Yield(Message):
    """
    A WAMP ``YIELD`` message.

    Formats:

    * ``[YIELD, INVOCATION.Request|id, Options|dict]``
    * ``[YIELD, INVOCATION.Request|id, Options|dict, Arguments|list]``
    * ``[YIELD, INVOCATION.Request|id, Options|dict, Arguments|list, ArgumentsKw|dict]``
    """

    MESSAGE_TYPE = 70
    """
   The WAMP message code for this type of message.
   """

    def __init__(self, request, args=None, kwargs=None, progress=None):
        """

        :param request: The WAMP request ID of the original call.
        :type request: int
        :param args: Positional values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type args: list or tuple or None
        :param kwargs: Keyword values for application-defined event payload.
           Must be serializable using any serializers in use.
        :type kwargs: dict or None
        :param progress: If ``True``, this result is a progressive invocation result, and subsequent
           results (or a final error) will follow.
        :type progress: bool or None
        """
        assert(type(request) in six.integer_types)
        assert(args is None or type(args) in [list, tuple])
        assert(kwargs is None or type(kwargs) == dict)
        assert(progress is None or type(progress) == bool)

        Message.__init__(self)
        self.request = request
        self.args = args
        self.kwargs = kwargs
        self.progress = progress

    @staticmethod
    def parse(wmsg):
        """
        Verifies and parses an unserialized raw message into an actual WAMP message instance.

        :param wmsg: The unserialized raw message.
        :type wmsg: list

        :returns: An instance of this class.
        """
        # this should already be verified by WampSerializer.unserialize
        ##
        assert(len(wmsg) > 0 and wmsg[0] == Yield.MESSAGE_TYPE)

        if len(wmsg) not in (3, 4, 5):
            raise ProtocolError("invalid message length {0} for YIELD".format(len(wmsg)))

        request = check_or_raise_id(wmsg[1], u"'request' in YIELD")
        options = check_or_raise_extra(wmsg[2], u"'options' in YIELD")

        args = None
        if len(wmsg) > 3:
            args = wmsg[3]
            if type(args) != list:
                raise ProtocolError("invalid type {0} for 'args' in YIELD".format(type(args)))

        kwargs = None
        if len(wmsg) > 4:
            kwargs = wmsg[4]
            if type(kwargs) != dict:
                raise ProtocolError("invalid type {0} for 'kwargs' in YIELD".format(type(kwargs)))

        progress = None

        if u'progress' in options:

            option_progress = options[u'progress']
            if type(option_progress) != bool:
                raise ProtocolError("invalid type {0} for 'progress' option in YIELD".format(type(option_progress)))

            progress = option_progress

        obj = Yield(request, args=args, kwargs=kwargs, progress=progress)

        return obj

    def marshal(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.marshal`
        """
        options = {}

        if self.progress is not None:
            options[u'progress'] = self.progress

        if self.kwargs:
            return [Yield.MESSAGE_TYPE, self.request, options, self.args, self.kwargs]
        elif self.args:
            return [Yield.MESSAGE_TYPE, self.request, options, self.args]
        else:
            return [Yield.MESSAGE_TYPE, self.request, options]

    def __str__(self):
        """
        Implements :func:`autobahn.wamp.interfaces.IMessage.__str__`
        """
        return "WAMP YIELD Message (request = {0}, args = {1}, kwargs = {2}, progress = {3})".format(self.request, self.args, self.kwargs, self.progress)
